import json
from pathlib import Path

import pytest

from quetz import channel_data
from quetz.tasks.indexing import update_indexes


@pytest.fixture
def empty_channeldata(dao):
    return channel_data.export(dao, "")


def test_update_indexes_empty_channel(config, public_channel, dao, empty_channeldata):
    pkgstore = config.get_package_store()

    update_indexes(dao, pkgstore, public_channel.name)

    files = pkgstore.list_files(public_channel.name)

    base_files = [
        'channeldata.json',
        'index.html',
        'noarch/index.html',
        'noarch/repodata.json',
    ]

    expected_files = base_files.copy()

    for suffix in ['.bz2', '.gz']:
        expected_files.extend(s + suffix for s in base_files)

    assert sorted(files) == sorted(expected_files)

    channel_dir = Path(pkgstore.channels_dir) / public_channel.name
    with open(channel_dir / 'channeldata.json', 'r') as fd:
        assert json.load(fd) == empty_channeldata


def test_update_indexes_empty_package(
    config, public_channel, public_package, dao, empty_channeldata
):
    pkgstore = config.get_package_store()

    update_indexes(dao, pkgstore, public_channel.name)

    files = pkgstore.list_files(public_channel.name)

    base_files = [
        'channeldata.json',
        'index.html',
        'noarch/index.html',
        'noarch/repodata.json',
    ]

    expected_files = base_files.copy()

    for suffix in ['.bz2', '.gz']:
        expected_files.extend(s + suffix for s in base_files)

    assert sorted(files) == sorted(expected_files)

    channel_dir = Path(pkgstore.channels_dir) / public_channel.name
    with open(channel_dir / 'channeldata.json', 'r') as fd:
        channeldata = json.load(fd)

    assert public_package.name in channeldata["packages"].keys()

    assert channeldata["packages"].pop(public_package.name) == {}
    assert channeldata == empty_channeldata


def test_update_indexes_with_package_version(
    config, public_channel, public_package, package_version, dao
):
    pkgstore = config.get_package_store()

    update_indexes(dao, pkgstore, public_channel.name)

    files = pkgstore.list_files(public_channel.name)

    base_files = [
        'channeldata.json',
        'index.html',
        'linux-64/index.html',
        'linux-64/repodata.json',
        'noarch/index.html',
        'noarch/repodata.json',
    ]

    expected_files = base_files.copy()

    for suffix in ['.bz2', '.gz']:
        expected_files.extend(s + suffix for s in base_files)

    expected_files.append(f"linux-64/{package_version.filename}")

    assert sorted(files) == sorted(expected_files)

    channel_dir = Path(pkgstore.channels_dir) / public_channel.name
    with open(channel_dir / 'channeldata.json', 'r') as fd:
        channeldata = json.load(fd)

    assert public_package.name in channeldata["packages"].keys()



import pytest

from quetz.authentication import auth_dao
from quetz.db_models import Email
from quetz.errors import ValidationError


def find_email(emails, e):
    for eo in emails:
        if eo.email == e:
            return eo
    raise KeyError("Missing email from list")


def test_get_user_by_identity_new_user(dao, config):
    profile = {
        "id": 4567,
        "login": "bartosz",
        "name": "bartosz",
        "avatar_url": "url",
    }
    provider = "github"

    user = auth_dao.get_user_by_identity(dao, provider, profile, config)

    assert user.username == 'bartosz'
    assert user.identities[0].provider == 'github'
    assert user.identities[0].identity_id == '4567'


def test_add_user_with_emails(dao, config):
    profile = {
        "id": 1234,
        "login": "wolfv",
        "name": "wolfv",
        "avatar_url": "url",
        "emails": [
            {"email": "w.vollprecht@abcdef.com", "verified": True, "primary": True},
            {
                "email": "wolf.vollprecht@lasersight.net",
                "verified": False,
                "primary": False,
            },
        ],
    }
    provider = "github"

    user = auth_dao.get_user_by_identity(
        dao, provider=provider, profile=profile, config=config
    )

    assert user.username == "wolfv"
    assert len(user.emails) == 1
    e = find_email(user.emails, "w.vollprecht@abcdef.com")
    assert e.email == "w.vollprecht@abcdef.com"
    assert e.verified is True
    assert e.primary is True

    profile["emails"][1]["verified"] = True

    user = auth_dao.get_user_by_identity(
        dao, provider=provider, profile=profile, config=config
    )

    e = find_email(user.emails, "wolf.vollprecht@lasersight.net")
    assert e.email == "wolf.vollprecht@lasersight.net"

    profile["emails"].append(
        {
            "email": "another.email@github.net",
            "verified": True,
            "primary": False,
        }
    )

    user = auth_dao.get_user_by_identity(
        dao, provider=provider, profile=profile, config=config
    )

    e = find_email(user.emails, "another.email@github.net")
    assert e.email == "another.email@github.net"

    profile["emails"] = profile["emails"][1:]
    profile["emails"][0]["primary"] = True
    profile["login"] = "franky"

    user = auth_dao.get_user_by_identity(
        dao, provider=provider, profile=profile, config=config
    )

    assert len(user.emails) == 2
    # assert user.username == "franky"
    e = find_email(user.emails, "wolf.vollprecht@lasersight.net")
    assert e.email == "wolf.vollprecht@lasersight.net"
    assert e.primary is True

    with pytest.raises(KeyError):
        e = find_email(user.emails, "w.vollprecht@abcdef.com")

    # Make sure that the email is properly deleted
    assert (
        dao.db.query(Email)
        .filter(Email.email == "w.vollprecht@abcdef.com")
        .one_or_none()
        is None
    )


def test_add_user_with_same_email(dao, config):
    profile = {
        "id": 1234,
        "login": "wolfv",
        "name": "wolfv",
        "avatar_url": "url",
        "emails": [
            {"email": "w.vollprecht@abcdef.com", "verified": True, "primary": True},
            {
                "email": "wolf.vollprecht@lasersight.net",
                "verified": True,
                "primary": False,
            },
        ],
    }
    provider = "github"

    profile2 = {
        "id": 1234,
        "login": "gitlab_wolfv",
        "name": "Wolf Vollprecht",
        "avatar_url": "url",
        "emails": [
            {
                "email": "wolf.vollprecht@lasersight.net",
                "verified": True,
                "primary": False,
            },
        ],
    }
    provider2 = "gitlab"

    user = auth_dao.get_user_by_identity(
        dao, provider=provider, profile=profile, config=config
    )

    assert user.username == "wolfv"
    e = find_email(user.emails, "w.vollprecht@abcdef.com")
    assert e.email == "w.vollprecht@abcdef.com"
    assert len(user.emails) == 2
    assert e.verified is True
    assert e.primary is True

    with pytest.raises(ValidationError):
        user = auth_dao.get_user_by_identity(
            dao, provider=provider2, profile=profile2, config=config
        )



"""add channel size limit

Revision ID: db1c56bf4d57
Revises: a3ffa287d074
Create Date: 2020-12-08 18:44:14.215836

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = 'db1c56bf4d57'
down_revision = 'a3ffa287d074'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('channels', sa.Column('size', sa.Integer(), nullable=True))
    op.add_column('channels', sa.Column('size_limit', sa.Integer(), nullable=True))
    op.add_column('package_versions', sa.Column('size', sa.Integer(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('package_versions', 'size')
    op.drop_column('channels', 'size_limit')
    op.drop_column('channels', 'size')
    # ### end Alembic commands ###



from quetz.dao import Dao
from quetz.pkgstores import PackageStore


def cleanup_channel_db(dao: Dao, channel_name: str, dry_run: bool):
    dao.cleanup_channel_db(channel_name, None, dry_run)


def cleanup_temp_files(pkgstore: PackageStore, channel_name: str, dry_run: bool):
    pkgstore.cleanup_temp_files(channel_name, dry_run)



"""adding download metrics

Revision ID: 8d1e9a9e0b1f
Revises: 794249a0b1bd
Create Date: 2020-12-18 12:29:29.994522

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = '8d1e9a9e0b1f'
down_revision = '794249a0b1bd'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    enum_values = ('hour', 'day', 'month', 'year')
    if op.get_context().dialect.name == 'postgresql':
        intervaltype = sa.dialects.postgresql.ENUM(
            *enum_values, name='intervaltype', create_type=True
        )
    else:
        intervaltype = sa.Enum(*enum_values, name='intervaltype')
    op.create_table(
        'package_version_metrics',
        sa.Column('package_version_id', sa.LargeBinary(length=16), nullable=False),
        sa.Column('metric_name', sa.String(length=255), nullable=False),
        sa.Column(
            'period',
            intervaltype,
            nullable=False,
        ),
        sa.Column('count', sa.Integer(), nullable=False, server_default=sa.text("0")),
        sa.Column('timestamp', sa.DateTime(), nullable=False),
        sa.ForeignKeyConstraint(
            ['package_version_id'],
            ['package_versions.id'],
        ),
        sa.PrimaryKeyConstraint(
            'package_version_id', 'metric_name', 'period', 'timestamp'
        ),
    )
    op.add_column(
        'package_versions', sa.Column('download_count', sa.Integer(), nullable=True)
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('package_versions', 'download_count')
    op.drop_table('package_version_metrics')
    if op.get_context().dialect.name == 'postgresql':
        op.execute("DROP TYPE intervaltype;")
    # ### end Alembic commands ###



import pytest

from quetz.database import sanitize_db_url


@pytest.mark.parametrize(
    "input_url,expected_output_url",
    (
        (
            "sqlite:///./quetz.sqlite",
            "sqlite:///./quetz.sqlite",
        ),  # No password, no effect
        (
            "postgresql+psycopg2://postgres_user:postgres_password@localhost:5432/postgres",  # noqa: E501
            "postgresql+psycopg2://postgres_user:***@localhost:5432/postgres",
        ),
        ("A:B@C:1111/DB", "A:***@C:1111/DB"),
        ("THISISNOTAURL", "THISISNOTAURL"),
    ),
)
def test_sanitize_db_url(input_url, expected_output_url):
    assert sanitize_db_url(input_url) == expected_output_url



import asyncio
import concurrent.futures
import inspect
import json
import logging
import pickle
import time
import uuid
from abc import abstractmethod
from multiprocessing import get_context
from typing import Callable, Dict, Optional, Union

from quetz.config import Config
from quetz.jobs.models import JobStatus, Task, TaskStatus

try:
    import redis  # type: ignore
    from rq import Queue

    rq_available = True
except ImportError:
    rq_available = False

logger = logging.getLogger("quetz.tasks")


def prepare_arguments(func: Callable, **resources):
    "select arguments for a function for resources based on its signature." ""

    # poorman's dependency injection pattern

    argnames = list(inspect.signature(func).parameters.keys())
    kwargs = {arg: value for arg, value in resources.items() if arg in argnames}

    return kwargs


def get_worker(config, num_procs=None):
    if config.configured_section("worker"):
        worker = config.worker_type
    else:
        worker = "thread"
    if worker == "thread":
        worker = ThreadingWorker(config)
    elif worker == "subprocess":
        worker = SubprocessWorker(config, executor_args={'max_workers': num_procs})
    elif worker == "redis":
        if rq_available:
            worker = RQManager(
                config.worker_redis_ip,
                config.worker_redis_port,
                config.worker_redis_db,
                config,
            )
        else:
            raise ValueError("redis and rq not installed on machine")
    else:
        raise ValueError("wrong configuration in worker.type")
    logger.debug(f"created worker of class {worker.__class__.__name__}")
    return worker


class WorkerProcess:
    def __init__(self, func, *args, **kwargs):
        self._exception = None
        self._connection = None
        self.func = func
        if isinstance(func, Callable):
            self._pickled_func = pickle.dumps(func)
        else:
            self._pickled_func = func
        self.func_args = args
        self.func_kwargs = kwargs

    @staticmethod
    def jobexecutor(conn, pickled_func, *args, **kwargs):
        """function executed in child process"""
        try:
            job_wrapper(pickled_func, *args, exc_passthrou=True, **kwargs)
        except Exception as exc:
            conn.send(exc)
            raise
        else:
            conn.send('ok')
        finally:
            conn.close()

    def fork_job(self):
        self.ctx = get_context("spawn")
        parent_conn, child_conn = self.ctx.Pipe(duplex=False)
        self._parent_conn = parent_conn

        self._process = self.ctx.Process(
            target=self.jobexecutor,
            args=(child_conn, self._pickled_func) + self.func_args,
            kwargs=self.func_kwargs,
        )
        self._process.start()

    def wait_for_job(self):
        exc = self._parent_conn.recv()
        self._process.join()
        if self._process.exitcode is None:
            raise RuntimeError
        if self._process.exitcode > 0:
            raise exc
        return

    def __call__(self):
        self.fork_job()
        self.wait_for_job()


def job_wrapper(
    func: Union[Callable, bytes],
    config,
    task_id=None,
    exc_passthrou=False,
    **kwargs,
):
    # database connections etc. are not serializable
    # so we need to recreate them in the process.
    # This allows us to manage database connectivity prior
    # to running a job.

    import logging
    import pickle

    from quetz.authorization import Rules
    from quetz.config import configure_logger
    from quetz.dao import Dao
    from quetz.database import get_session
    from quetz.deps import get_remote_session

    configure_logger(config)

    logger = logging.getLogger("quetz.worker")

    pkgstore = kwargs.pop("pkgstore", None)
    db = kwargs.pop("db", None)
    dao = kwargs.pop("dao", None)
    auth = kwargs.pop("auth", None)
    session = kwargs.pop("session", None)

    if db:
        close_session = False
    elif dao:
        db = dao.db
        close_session = False
    else:
        db = get_session(config.sqlalchemy_database_url)
        close_session = True

    user_id: Optional[str]
    if task_id:
        task = db.query(Task).filter(Task.id == task_id).one_or_none()
        if not task:
            raise KeyError(f"Task '{task_id}' not found")
        # take extra arguments from job definition
        if task.job.extra_args:
            job_extra_args = json.loads(task.job.extra_args)
            kwargs.update(job_extra_args)
        if task.job.owner_id:
            user_id = str(uuid.UUID(bytes=task.job.owner_id))
        else:
            user_id = None
    else:
        task = None
        user_id = None

    if not pkgstore:
        pkgstore = config.get_package_store()

    dao = Dao(db)

    if not auth:
        browser_session: Dict[str, str] = {}
        api_key = None
        if user_id:
            browser_session['user_id'] = user_id
        auth = Rules(api_key, browser_session, db)
    if not session:
        session = get_remote_session()

    if task:
        task.status = TaskStatus.running
        task.job.status = JobStatus.running
        db.commit()

    callable_f: Callable = pickle.loads(func) if isinstance(func, bytes) else func

    extra_kwargs = prepare_arguments(
        callable_f,
        dao=dao,
        auth=auth,
        session=session,
        config=config,
        pkgstore=pkgstore,
        user_id=user_id,
    )

    kwargs.update(extra_kwargs)

    try:
        callable_f(**kwargs)
    except Exception as exc:
        if task:
            task.status = TaskStatus.failed
        logger.error(
            f"exception occurred when evaluating function {callable_f.__name__}:{exc}"
        )
        if exc_passthrou:
            raise exc
    else:
        if task:
            task.status = TaskStatus.success
    finally:
        db.commit()
        if close_session:
            db.close()


class AbstractWorker:
    @abstractmethod
    def execute(self, func, **kwargs):
        """execute function func on the worker."""

    @abstractmethod
    def wait(self):
        """wait for all jobs to finish"""


class AbstractJob:
    """Single job (function call)"""

    @property
    @abstractmethod
    def done(self):
        """job status"""


class FutureJob(AbstractJob):
    def __init__(self, future: concurrent.futures.Future):
        self._future = future

    @property
    def status(self):
        is_running = self._future.running()
        if is_running:
            return "running"

        completed = self._future.done()
        if completed:
            if self._future.exception():
                return "failed"
            else:
                return "success"
        return "pending"

    @property
    def done(self):
        return self.status in ["failed", "success"]

    async def wait(self, waittime=0.1):
        while not self.done:
            await asyncio.sleep(waittime)
        if self.status == "failed":
            raise self._future.exception()


class PoolExecutorWorker(AbstractWorker):
    _executor = None
    executor_cls: type

    def __init__(
        self,
        config: Config,
        executor_args: dict = {},
    ):
        if self._executor is None:
            logger.debug(f"creating a new executor of class {self.executor_cls}")
            type(self)._executor = self.executor_cls(**executor_args)

        self.config = config
        self.future = None

    def execute(self, func, *args, **kwargs):
        self.future = self._executor.submit(
            job_wrapper, func, self.config, *args, **kwargs
        )
        return FutureJob(self.future)

    async def wait(self):
        "helper function mainly used in tests"
        loop = asyncio.get_event_loop()
        if self.future:
            return await loop.run_in_executor(None, self.future.result)


class ThreadingWorker(PoolExecutorWorker):
    """simple worker that runs jobs in threads"""

    executor_cls = concurrent.futures.ThreadPoolExecutor


class SubprocessWorker(PoolExecutorWorker):
    """subprocess worker runs each job in a subprocess and destroys it after
    the job is finished to release resources
    """

    executor_cls = concurrent.futures.ProcessPoolExecutor

    def __init__(self, config: Config, executor_args: dict = {}):
        if 'max_workers' not in executor_args:
            executor_args['max_workers'] = 2
        super().__init__(config, executor_args)

    def execute(self, func, *args, **kwargs):
        process = WorkerProcess(func, self.config, *args, **kwargs)
        self.future = self._executor.submit(process)
        return FutureJob(self.future)


class RQManager(AbstractWorker):
    def __init__(
        self,
        host,
        port,
        db,
        config: Config,
        no_testing=True,
    ):
        self.host = host
        self.port = port
        self.db = db
        self.config = config
        self.conn = redis.StrictRedis(host=self.host, port=self.port, db=self.db)
        self.queue = Queue(connection=self.conn, is_async=no_testing)

    def execute(self, func, *args, **kwargs):
        self.job = self.queue.enqueue(
            job_wrapper,
            func,
            self.config,
            *args,
            **kwargs,
        )

    # the function is blocking in nature and is declared
    # as 'async' so as to make redis-queue compatible
    # with the testing framework. It is not to be used otherwise.
    async def wait(self):
        while not self.job.is_finished:
            time.sleep(1)
        if self.job.result:
            return self.job.result



import json
import logging
import os
import time
import uuid
from concurrent.futures import ThreadPoolExecutor, as_completed

from sqlalchemy.exc import IntegrityError

from quetz import authorization, rest_models
from quetz.condainfo import CondaInfo
from quetz.config import Config
from quetz.dao import Dao
from quetz.exceptions import PackageError

from .indexing import update_indexes

logger = logging.getLogger("quetz.tasks")


def uuid_to_bytes(id):
    if isinstance(id, str):
        id = uuid.UUID(id).bytes
    return id


def handle_condainfo(pkgstore, channel_name, fname):
    """Fetch CondaInfo for a package from pkgstaore"""

    fid = pkgstore.serve_path(channel_name, fname)

    try:
        condainfo = CondaInfo(fid, fname, lazy=False)
    except PackageError:
        logger.error(f"Package {fname} is not a tar.bzip2 file")
        condainfo = None

    return condainfo


def handle_file(channel_name, condainfo, dao, user_id):
    """Add or update conda package info to database"""

    filename = os.path.split(condainfo._filename)[-1]
    package_name = condainfo.info["name"]
    package_format = condainfo.package_format
    platform = condainfo.info["subdir"]
    version = condainfo.info["version"]
    build_number = condainfo.info["build_number"]
    build_string = condainfo.info["build"]
    size = condainfo.info["size"]
    info = json.dumps(condainfo.info)

    package = dao.get_package(channel_name, package_name)

    if not package:
        logger.debug(f"Creating package {package_name}")

        dao.create_package(
            channel_name,
            rest_models.Package(
                name=package_name,
                summary=condainfo.about.get("summary", "n/a"),
                description=condainfo.about.get("description", "n/a"),
            ),
            user_id,
            authorization.OWNER,
        )

        dao.update_package_channeldata(
            channel_name, package_name, condainfo.channeldata
        )

    try:
        logger.debug(
            f"Adding package {channel_name}/{platform}/{package_name}"
            + f"-{version}-{build_string}"
        )
        dao.create_version(
            channel_name=channel_name,
            package_name=package_name,
            package_format=package_format,
            platform=platform,
            version=version,
            build_number=build_number,
            build_string=build_string,
            size=size,
            filename=filename,
            info=info,
            uploader_id=user_id,
            upsert=True,
        )
        dao.db.commit()
    except IntegrityError:
        dao.rollback()
        logger.error(
            f"Duplicate package {channel_name}/{package_name}"
            + f"-{condainfo.info['version']}"
        )
    dao.db.commit()


def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""

    for i in range(0, len(lst), n):
        yield lst[i : i + n]


def reindex_packages_from_store(
    dao: Dao, config: Config, channel_name: str, user_id, sync: bool = True
):
    """Reindex packages from files in the package store"""

    logger.debug(f"Re-indexing channel {channel_name}")

    channel = dao.get_channel(channel_name)
    pkg_db = []
    if channel:
        if not sync:
            for package in channel.packages:
                dao.db.delete(package)
            dao.db.commit()
        else:
            dao.cleanup_channel_db(channel_name)
            for package in channel.packages:
                for pv in package.package_versions:  # type: ignore
                    pkg_db.append(f"{pv.platform}/{pv.filename}")
            dao.db.commit()
    else:
        data = rest_models.Channel(
            name=channel_name, description="re-indexed from store", private=True
        )
        channel = dao.create_channel(data, user_id, authorization.OWNER)

    logger.debug(f"Reading package list for channel {channel_name}")
    user_id = uuid_to_bytes(user_id)
    pkgstore = config.get_package_store()
    all_files = pkgstore.list_files(channel_name)
    pkg_files = [f for f in all_files if f.endswith(".tar.bz2")]
    nthreads = config.general_package_unpack_threads

    logger.debug(f"Found {len(pkg_db)} packages for channel {channel_name} in database")
    logger.debug(
        f"Found {len(pkg_files)} packages for channel {channel_name} in pkgstore"
    )

    pkg_files = list(set(pkg_files) - set(pkg_db))
    logger.debug(
        f"Importing {len(pkg_files)} packages for channel {channel_name}"
        + " from pkgstore"
    )

    for pkg_group in chunks(pkg_files, nthreads * 8):
        tic = time.perf_counter()
        with ThreadPoolExecutor(max_workers=nthreads) as executor:
            results = []
            for fname in pkg_group:
                results.append(
                    executor.submit(handle_condainfo, pkgstore, channel_name, fname)
                )
            for future in as_completed(results):
                condainfo = future.result()
                if condainfo:
                    handle_file(channel_name, condainfo, dao, user_id)

        toc = time.perf_counter()
        logger.debug(
            f"Imported files {pkg_group[0]} to {pkg_group[-1]} "
            + f"for channel {channel_name} in {toc - tic:0.4f} seconds "
            + f"using {nthreads} threads"
        )

        try:
            update_indexes(dao, pkgstore, channel_name)
            dao.db.commit()
        except IntegrityError:
            dao.rollback()
            logger.error(f"Update index {channel_name} failed")
    dao.cleanup_channel_db(channel_name)
    dao.db.commit()



import logging
from datetime import datetime
from typing import Optional

import requests

from quetz.dao import Dao


def synchronize_metrics_from_mirrors(
    channel_name: str,
    dao: Dao,
    session: requests.Session,
    now: datetime = datetime.utcnow(),
):
    logger = logging.getLogger("quetz")
    channel = dao.get_channel(channel_name)
    if not channel:
        return
    for m in channel.mirrors:
        if not m.metrics_endpoint:
            logger.warning(
                f"metrics endpoint not configured for mirror {m.url}."
                "Skipping metrics synchronisation"
            )
            continue
        query_str = ["period=H"]
        start_time: Optional[datetime]
        if m.last_synchronised:
            start_time = m.last_synchronised.replace(minute=0, second=0, microsecond=0)
        else:
            start_time = None
        if isinstance(start_time, datetime):
            query_str.append(f"start={start_time.isoformat()}")

        # exclude incomplete intervals (the current hour)
        end_time = now.replace(minute=0, second=0, microsecond=0)

        if start_time == end_time:
            logger.debug(f"metrics data for mirror {m.url} are up-to-date")
            continue

        query_str.append(f"end={end_time.isoformat()}")

        metrics_url = m.metrics_endpoint + "?" + "&".join(query_str)
        response = session.get(metrics_url)

        if response.status_code != 200:
            logger.error(
                f"mirror server {metrics_url} returned bad response with code "
                f"{response.status_code} and message {response.text}"
            )
            continue

        response_data = response.json()
        try:
            packages = response_data["packages"]
        except KeyError:
            logger.error(
                f"malfromated respose received from {metrics_url}: "
                "missing 'packages' key"
            )
            continue

        for platform_filename, data in packages.items():
            platform, filename = platform_filename.split('/')
            for s in data["series"]:
                timestamp = datetime.fromisoformat(s["timestamp"])
                count = s["count"]
                dao.incr_download_count(
                    channel_name, filename, platform, timestamp, count
                )
        logger.debug(f"synchronized metrics from {metrics_url}")
        m.last_synchronised = end_time
        dao.db.commit()



# Copyright 2020 QuantStack
# Distributed under the terms of the Modified BSD License.

import enum
import uuid
from datetime import date
from typing import Optional

from fastapi import HTTPException, status
from sqlalchemy import or_
from sqlalchemy.orm import Session

import quetz.config

from .db_models import ApiKey, ChannelMember, PackageMember, User

OWNER = "owner"
MAINTAINER = "maintainer"
MEMBER = "member"

SERVER_OWNER = OWNER
SERVER_MAINTAINER = MAINTAINER
SERVER_MEMBER = MEMBER
SERVER_USER = None

ROLES = [OWNER, MAINTAINER, MEMBER]


class ServerRole(str, enum.Enum):
    OWNER = OWNER
    MAINTAINER = MAINTAINER
    MEMBER = MEMBER
    USER = None


class Rules:
    def __init__(self, API_key: Optional[str], session: dict, db: Session):
        self.API_key = API_key
        self.session = session
        self.db = db

    def get_valid_api_key(self) -> Optional[ApiKey]:
        if not self.API_key:
            return None
        return (
            self.db.query(ApiKey)
            .filter(ApiKey.key == self.API_key, ~ApiKey.deleted)
            .filter(
                ApiKey.key == self.API_key,
                or_(ApiKey.expire_at >= date.today(), ApiKey.expire_at.is_(None)),
            )
            .one_or_none()
        )

    def get_owner(self) -> Optional[bytes]:
        """
        gets the id of the owner of the authenticated session, if any. Either:
          a) the owner of the API key that is used for authentication
          b) the user_id of the encrypted session cookie
        """
        owner_id = None

        if self.API_key:
            api_key = self.get_valid_api_key()
            if api_key:
                owner_id = api_key.owner_id
        else:
            user_id = self.session.get("user_id")
            if user_id:
                owner_id = uuid.UUID(user_id).bytes

        return owner_id

    def assert_owner(self) -> bytes:
        owner_id = self.get_owner()

        if not owner_id or not self.db.query(User).filter(User.id == owner_id).count():
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Not logged in",
            )

        return owner_id

    def get_user(self) -> Optional[bytes]:
        """
        gets the id of the user of the authenticated session, if any. Either:
          a) the user_id of the API key (not its owner!) that is used for authentication
          b) the user_id of the encrypted session cookie
        """
        user_id = None

        if self.API_key:
            api_key = self.get_valid_api_key()
            if api_key:
                user_id = api_key.user_id
        else:
            user_id = self.session.get("user_id")
            if user_id:
                user_id = uuid.UUID(user_id).bytes

        return user_id

    def assert_user(self) -> bytes:
        user_id = self.get_user()

        if not user_id or not self.db.query(User).filter(User.id == user_id).count():
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Not logged in",
            )

        return user_id

    def assert_read_user_data(self, requested_user_id: bytes):
        user_id = self.assert_user()

        if not (requested_user_id == user_id):
            self.assert_server_roles([SERVER_OWNER, SERVER_MAINTAINER])

        return user_id

    def assert_delete_user(self, requested_user_id: bytes):
        user_id = self.assert_user()

        if not (requested_user_id == user_id):
            self.assert_server_roles([SERVER_OWNER, SERVER_MAINTAINER])

        return user_id

    def assert_assign_user_role(self, role: Optional[str]):
        if role == SERVER_MAINTAINER or role == SERVER_OWNER:
            return self.assert_server_roles([SERVER_OWNER])
        if role == SERVER_MEMBER:
            return self.assert_server_roles([SERVER_OWNER, SERVER_MAINTAINER])

    def assert_server_roles(self, roles: list, msg: Optional[str] = None):
        user_id = self.assert_user()

        if not self.has_server_roles(user_id, roles):
            detail = (
                msg or "this operation requires " + " or ".join(roles) + " roles",
            )

            raise HTTPException(status_code=status.HTTP_403_FORBIDDEN, detail=detail)

        return user_id

    def has_server_roles(self, user_id, roles: list):
        pm = quetz.config.get_plugin_manager()
        res = self.db.query(User).filter(User.id == user_id).one_or_none()

        if res:
            user_role = res.role
            if user_role in roles:
                permissions_check = pm.hook.check_additional_permissions(
                    db=self.db, user_id=user_id, user_role=user_role
                )
                if len(permissions_check):
                    if all(permissions_check):
                        return res
                    else:
                        return None
                else:
                    return res
            else:
                return None
        else:
            return None

    def has_channel_role(self, user_id: bytes, channel_name: str, roles: list):
        return (
            self.db.query(ChannelMember)
            .filter(ChannelMember.user_id == user_id)
            .filter(ChannelMember.channel_name == channel_name)
            .filter(ChannelMember.role.in_(roles))
            .one_or_none()
        )

    def has_package_role(
        self, user_id, channel_name: str, package_name: str, roles: list
    ):
        return (
            self.db.query(PackageMember)
            .filter(PackageMember.user_id == user_id)
            .filter(PackageMember.channel_name == channel_name)
            .filter(PackageMember.package_name == package_name)
            .filter(PackageMember.role.in_(roles))
            .one_or_none()
        )

    def has_channel_or_package_roles(
        self,
        user_id,
        channel_name: str,
        channel_roles: list,
        package_name: str,
        package_roles: list,
    ):
        return (
            self.is_user_elevated(user_id)
            or self.has_channel_role(user_id, channel_name, channel_roles)
            or self.has_package_role(user_id, channel_name, package_name, package_roles)
        )

    def is_user_elevated(self, user_id):
        return self.has_server_roles(user_id, [SERVER_OWNER, SERVER_MAINTAINER])

    def assert_channel_roles(self, channel_name: str, channel_roles: list):
        user_id = self.assert_user()

        if not (
            self.is_user_elevated(user_id)
            or self.has_channel_role(user_id, channel_name, channel_roles)
        ):
            raise HTTPException(
                status_code=status.HTTP_403_FORBIDDEN, detail="No permission"
            )

    def assert_channel_or_package_roles(
        self,
        channel_name: str,
        channel_roles: list,
        package_name: str,
        package_roles: list,
    ):
        user_id = self.assert_user()

        if not self.has_channel_or_package_roles(
            user_id, channel_name, channel_roles, package_name, package_roles
        ):
            raise HTTPException(
                status_code=status.HTTP_403_FORBIDDEN, detail="No permission"
            )

    def assert_add_channel_member(self, channel_name: str, role: str):
        self.assert_channel_roles(channel_name, [OWNER])

    def assert_remove_channel_member(self, channel_name: str, role: str):
        self.assert_channel_roles(channel_name, [OWNER])

    def assert_add_package_member(self, channel_name, package_name, role):
        self.assert_channel_or_package_roles(
            channel_name, [OWNER, MAINTAINER], package_name, [OWNER]
        )

    def assert_create_api_key_roles(self, roles):
        for role in roles:
            if role.package:
                required_package_role = (
                    [OWNER] if role.role == OWNER else [OWNER, MAINTAINER]
                )
                self.assert_channel_or_package_roles(
                    role.channel,
                    [OWNER, MAINTAINER],
                    role.package,
                    required_package_role,
                )
            elif role.channel:
                required_channel_roles = (
                    [OWNER] if role.role == OWNER else [OWNER, MAINTAINER]
                )
                self.assert_channel_roles(role.channel, required_channel_roles)
            else:
                # create key without assigning special channel/package privilages
                return True

    def assert_delete_api_key(self, api_key):
        user_id = self.assert_user()

        if (
            not self.is_user_elevated(user_id)
            and not api_key.user_id == user_id
            and not api_key.owner_id == user_id
        ):
            raise HTTPException(
                status_code=status.HTTP_403_FORBIDDEN, detail="No permission"
            )

    def assert_upload_file(self, channel_name, package_name):
        self.assert_channel_or_package_roles(
            channel_name, [OWNER, MAINTAINER], package_name, [OWNER, MAINTAINER]
        )

    def assert_create_mirror_channel(self):
        self.assert_server_roles([SERVER_OWNER, SERVER_MAINTAINER])

    def assert_create_channel(self):
        self.assert_server_roles([SERVER_OWNER, SERVER_MAINTAINER, SERVER_MEMBER])

    def assert_update_channel_info(self, channel_name: str):
        self.assert_channel_roles(channel_name, [OWNER, MAINTAINER])

    def assert_register_mirror(self, channel_name: str):
        self.assert_channel_roles(channel_name, [OWNER, MAINTAINER])

    def assert_unregister_mirror(self, channel_name: str):
        self.assert_register_mirror(channel_name)

    def assert_create_package(self, channel_name: str):
        self.assert_channel_roles(channel_name, [OWNER, MAINTAINER])

    def assert_create_proxy_channel(self):
        self.assert_server_roles([SERVER_OWNER, SERVER_MAINTAINER])

    def assert_list_channel_members(self, channel_name: str):
        self.assert_channel_roles(channel_name, [OWNER, MAINTAINER])

    def assert_synchronize_mirror(self, channel_name):
        self.assert_channel_roles(channel_name, [OWNER, MAINTAINER])

    def assert_reindex_channel(self, channel_name):
        self.assert_channel_roles(channel_name, [OWNER, MAINTAINER])

    def assert_validate_package_cache(self, channel_name):
        self.assert_server_roles([SERVER_OWNER, SERVER_MAINTAINER])

    def assert_overwrite_package_version(self, channel_name, package_name):
        self.assert_channel_or_package_roles(
            channel_name, [OWNER], package_name, [OWNER]
        )

    def assert_delete_channel(self, channel):
        self.assert_channel_roles(channel.name, [OWNER, MAINTAINER])

    def assert_channel_read(self, channel):
        if channel.private:
            self.assert_channel_roles(channel.name, [OWNER, MAINTAINER, MEMBER])

    def assert_set_channel_size_limit(self):
        self.assert_server_roles(
            [SERVER_OWNER, SERVER_MAINTAINER],
            msg="only server maintainer or owner can set channel size limit",
        )

    def assert_channel_db_cleanup(self, channel_name):
        self.assert_channel_roles(channel_name, [OWNER])

    def assert_package_read(self, package):
        if package.channel.private:
            self.assert_channel_or_package_roles(
                package.channel_name,
                [OWNER, MAINTAINER, MEMBER],
                package.name,
                [OWNER, MAINTAINER, MEMBER],
            )

    def assert_package_write(self, package):
        self.assert_channel_or_package_roles(
            package.channel_name, [OWNER, MAINTAINER], package.name, [OWNER, MAINTAINER]
        )

    def assert_package_delete(self, package):
        self.assert_channel_or_package_roles(
            package.channel_name, [OWNER, MAINTAINER], package.name, [OWNER, MAINTAINER]
        )

    def assert_jobs(self, owner_id: Optional[bytes] = None):
        user_id = self.assert_user()
        if not self.is_user_elevated(user_id):
            if not owner_id or user_id != owner_id:
                raise HTTPException(
                    status_code=status.HTTP_401_UNAUTHORIZED,
                    detail="Not allowed",
                )



"""adding function args to job spec

Revision ID: 0a0ab48887ab
Revises: 3c3288034362
Create Date: 2021-02-24 16:58:56.886842

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = '0a0ab48887ab'
down_revision = '3c3288034362'
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table('jobs', schema=None) as batch_op:
        batch_op.add_column(sa.Column('extra_args', sa.String(), nullable=True))


def downgrade():
    with op.batch_alter_table('jobs', schema=None) as batch_op:
        batch_op.drop_column('extra_args')



import hashlib
import time

import pytest
from authlib.jose import JsonWebKey, jwt
from authlib.oidc.core.util import create_half_hash
from fastapi.testclient import TestClient

from quetz.authentication import github as auth_github
from quetz.authentication import google as auth_google
from quetz.authentication.azuread import AzureADAuthenticator
from quetz.authentication.jupyterhub import JupyterhubAuthenticator
from quetz.authorization import SERVER_OWNER
from quetz.dao import Dao
from quetz.db_models import Channel, ChannelMember, Identity, Profile, User
from quetz.testing.utils import AsyncPathMapDispatch


@pytest.fixture
def default_role():
    return "member"


@pytest.fixture
def user_group():
    return None


@pytest.fixture
def user_roles(provider_spec, login, user_group):
    # add user login identified with provider provider_spec to
    # group user_group
    provider_name, _ = provider_spec
    roles = {"admins": [], "members": [], "maintainers": []}

    if user_group:
        roles[user_group].append(f"{provider_name}:{login}")
    return roles


@pytest.fixture
def config_extra(default_role, user_roles):
    config_values = ["[users]"]
    if default_role:
        config_values.append(f'default_role = "{default_role}"')
    for group in ['admins', 'members', 'maintainers']:
        group_users = user_roles.get(group, [])
        config_values.append(f'{group} = {group_users}')
    return "\n".join(config_values)


@pytest.fixture
def login():
    return "user-with-role"


@pytest.fixture
def github_response(login):
    response = {
        '/login/oauth/access_token': {'body': {'access_token': 'b'}},
        '/user': {
            "body": {
                "login": login,
                "avatar_url": "",
                "id": login + "_id",
                "name": "monalisa",
            }
        },
    }

    return "test_github", response


@pytest.fixture
def azuread_response(login):
    response = {
        '/3cd6563a-bbcf-4e30-bbc6-ea3445c75c79/oauth2/v2.0/token': {
            'body': {'access_token': 'b'}
        },
        '/oidc/userinfo': {
            "body": {
                'sub': login + '_id',
                'name': 'monalisa',
                'family_name': 'Lisa',
                'given_name': 'Mona',
                'picture': 'https://graph.microsoft.com/v1.0/me/photo/$value',
                'email': login,
            }
        },
        '/3cd6563a-bbcf-4e30-bbc6-ea3445c75c79/v2.0/.well-known/openid-configuration': {
            "body": {
                "token_endpoint": "https://login.microsoftonline.com/"
                "3cd6563a-bbcf-4e30-bbc6-ea3445c75c79/oauth2/v2.0/token",
                "token_endpoint_auth_methods_supported": [
                    "client_secret_post",
                    "private_key_jwt",
                    "client_secret_basic",
                ],
                "jwks_uri": "https://login.microsoftonline.com/"
                "3cd6563a-bbcf-4e30-bbc6-ea3445c75c79/discovery/v2.0/keys",
                "response_modes_supported": ["query", "fragment", "form_post"],
                "subject_types_supported": ["pairwise"],
                "id_token_signing_alg_values_supported": ["RS256"],
                "response_types_supported": [
                    "code",
                    "id_token",
                    "code id_token",
                    "id_token token",
                ],
                "scopes_supported": ["openid", "profile", "email", "offline_access"],
                "issuer": "https://login.microsoftonline.com/"
                "3cd6563a-bbcf-4e30-bbc6-ea3445c75c79/v2.0",
                "request_uri_parameter_supported": False,
                "userinfo_endpoint": "https://graph.microsoft.com/oidc/userinfo",
                "authorization_endpoint": "https://login.microsoftonline.com/"
                "3cd6563a-bbcf-4e30-bbc6-ea3445c75c79/oauth2/v2.0/authorize",
                "device_authorization_endpoint": "https://login.microsoftonline.com/"
                "3cd6563a-bbcf-4e30-bbc6-ea3445c75c79/oauth2/v2.0/devicecode",
                "http_logout_supported": True,
                "frontchannel_logout_supported": True,
                "end_session_endpoint": "https://login.microsoftonline.com/"
                "3cd6563a-bbcf-4e30-bbc6-ea3445c75c79/oauth2/v2.0/logout",
                "claims_supported": [
                    "sub",
                    "iss",
                    "cloud_instance_name",
                    "cloud_instance_host_name",
                    "cloud_graph_host_name",
                    "msgraph_host",
                    "aud",
                    "exp",
                    "iat",
                    "auth_time",
                    "acr",
                    "nonce",
                    "preferred_username",
                    "name",
                    "tid",
                    "ver",
                    "at_hash",
                    "c_hash",
                    "email",
                ],
                "kerberos_endpoint": "https://login.microsoftonline.com/"
                "3cd6563a-bbcf-4e30-bbc6-ea3445c75c79/kerberos",
                "tenant_region_scope": "EU",
                "cloud_instance_name": "microsoftonline.com",
                "cloud_graph_host_name": "graph.windows.net",
                "msgraph_host": "graph.microsoft.com",
                "rbac_url": "https://pas.windows.net",
            }
        },
    }

    return "test_azuread", response


@pytest.fixture
def jupyter_response(login):
    response = {
        '/hub/api/oauth2/token': {
            'body': {'access_token': 'b', 'token_type': 'Bearer'}
        },
        '/hub/api/authorizations/token/b': {
            "body": {
                "name": login,
            }
        },
    }

    return "test_jupyter", response


GOOGLE_CERT = """-----BEGIN CERTIFICATE-----
MIIDazCCAlOgAwIBAgIUHFhnnOpbCrKVxDGjpRBhXMolH/kwDQYJKoZIhvcNAQEL
BQAwRTELMAkGA1UEBhMCQVUxEzARBgNVBAgMClNvbWUtU3RhdGUxITAfBgNVBAoM
GEludGVybmV0IFdpZGdpdHMgUHR5IEx0ZDAeFw0yMTAxMTgxNDI4NTJaFw0yNDAx
MTgxNDI4NTJaMEUxCzAJBgNVBAYTAkFVMRMwEQYDVQQIDApTb21lLVN0YXRlMSEw
HwYDVQQKDBhJbnRlcm5ldCBXaWRnaXRzIFB0eSBMdGQwggEiMA0GCSqGSIb3DQEB
AQUAA4IBDwAwggEKAoIBAQDH/ZUbewS+Tukx50UjRZ4bWij1tzoaQTo1NDWCwv7S
OGKY083dFR567aag1rhUHjiniuf8UVv935Ydq4VWmwV6N6XHoSLNXQSHcAN7VFos
We5ENAfBz4JmfZnZfG+QBB0nGtd7hjE/xtOUsOhuzposi8fP+FXwPMVMmgfuWS6Z
1SNODuwmtwobhe7x8ez6l70lWUIhac5SAQwUicsdqlH1gBFkcGMmFRj9DrslpYvu
8bq1UcGE3NhzHkIo4ssAE2NlHtzbDnxYpXGC3aEPtWHL/mFVneud2a6ZcEVDc1Nq
cwcWliUwK6FI/wVcNYbCX6AozSor2+XCtUTyXdGoj+spAgMBAAGjUzBRMB0GA1Ud
DgQWBBRV8o2cbHAis+yj0dzP1j25bCOjkjAfBgNVHSMEGDAWgBRV8o2cbHAis+yj
0dzP1j25bCOjkjAPBgNVHRMBAf8EBTADAQH/MA0GCSqGSIb3DQEBCwUAA4IBAQAA
8fT4OiTn1zPBBV702LP3cxJd+iOtfTxBFTOB9QATYpKmaQvgf6gRLHgzgFJSpn0e
e/NH9wEllLqwbZbmz9kOH/LxaY8WD8eWbWTR0+2dcaxG0qjitlijnHfEsVANvv9O
FK+exD3aCdMK/WUr5Shae2jRhDzLcGYdMeCD9Nxc1ShmHDA/4eUqk+SOso+O5v6i
ZqmzBWk/u9Z7JvQ41R0OG1tJoNrD02ctH2lel/ZX/7Ff4HxK6QyzBNtEnvJBXy/R
3qXwz4/xMdlqU0GKGpYLBvm180Nvmuohdtvzt8A8/nUKyNDJwzLHp7BjoyK9JxYy
xT/6TQKFDnCKt1Mzg1Py
-----END CERTIFICATE-----"""

GOOGLE_PRIVATE_KEY = """-----BEGIN RSA PRIVATE KEY-----
MIIEpAIBAAKCAQEAx/2VG3sEvk7pMedFI0WeG1oo9bc6GkE6NTQ1gsL+0jhimNPN
3RUeeu2moNa4VB44p4rn/FFb/d+WHauFVpsFejelx6EizV0Eh3ADe1RaLFnuRDQH
wc+CZn2Z2XxvkAQdJxrXe4YxP8bTlLDobs6aLIvHz/hV8DzFTJoH7lkumdUjTg7s
JrcKG4Xu8fHs+pe9JVlCIWnOUgEMFInLHapR9YARZHBjJhUY/Q67JaWL7vG6tVHB
hNzYcx5CKOLLABNjZR7c2w58WKVxgt2hD7Vhy/5hVZ3rndmumXBFQ3NTanMHFpYl
MCuhSP8FXDWGwl+gKM0qK9vlwrVE8l3RqI/rKQIDAQABAoIBAB48JDLHYmwzGeZF
hJpUiBayhsa/MLWPbvFkN0LRoBzAEYfxXYozCyyiiTJ/w9ZTy1TpFzF6S2IST2uk
5r+1KBrWFuYbYluR2IFxWdVnZ0qVPgRpqVKPwLMmAgBzY5puRMoIsNMn8oIl2Q79
v+YgrgZWC5tRfAyZ42o1T0WljfoLjqImry1gEFpfmKXmoKbcJhd0fTfobuG6lxY6
fkySEDeYuq4t15bjSBzqgzLD7fqSRS55yukUpSucQETB/grtLykVdHFAXP/KHWy3
06Mj3oJAciHTwmIKV0+9QgTU77bYAupLaRX6/pgpLSXt9XUwbI0/I+xq8Fo8RBrt
uxtCFgECgYEA5cideXxexpc3zXjg3dkC9x4ktVh4LHyQp/Ua+kRQDvHtf7E/44Bd
GDpQnxrJ3jfhyXfdLshlI8EouT7STyC0cjtmME5WhleGBokAvdp8EEuGpArCR8iK
1uKiFzUvszv2brr50K6s6/YPaRIGbQTFLfs7hZQYXjabx2PXwHmqJ1kCgYEA3s7J
x07mHqkz56k/MONyVUhaW+PFvUTqolGe3D3nULd7XBbOX3yCzZjRVHBfBKv/Tyqs
CWl9BZy8SBZX7ieiEO+5xX2bypfUxGeWznYJoYs+RDW5vEfnxFVvJ7KD/8lqq4WC
0YY/PGziCah2iCQumypFB51FUXbljEDcGxROOFECgYEAkxu1nYI/FvrW0efyZnU5
jcWxkJv8C9cPsUedJt43Nuoxt49drKOQdiNXXBUFagvytE3Vv86x2YsfLEGI2PnC
LGPUz1ZH1KgR+PsbC3Dl/nSr1TfCG7zLDjl3tk3ppODdqxRvPOenc0VLpmPQ01i7
d+2gtKsUUrS5VJSaGvKJObkCgYEAzuWlV6+7XvNuYIu4QzSiAfGa/sNG5tetLieu
5gOR3lFTexMudlrPuA1VLRzgDx2Mij4s3NyZHPILoMEmy97/zsxdbLeUSI+vIuay
kmvny5vaqUpefCklXhqbinhpvMeTh00GSnxoEjtltuQ5lXhL0whwa36uVNScmh3M
hlTXwdECgYAwqpJKC2Ls7i+3lSPomCNK1VtbVZzfHl1YrNy7y2Eo8DsFkcnHWZJ+
qPCpHZ4sbgfziCRxOaGa+1kOXjhgXITPA/tkkhh4cglOvhHNBEVzdfHBMDxZd3HF
mD1rl6xev7GRoqUYdKYdt9NJyDGEULZ6NbIWyXo3kTp7HdQLRn0BJg==
-----END RSA PRIVATE KEY-----
"""


@pytest.mark.skipif(
    not hasattr(hashlib, "RS256"), reason="hashlib does not support RS256"
)
@pytest.fixture
def google_response(login):
    google_client_id = 'aaa'
    key = GOOGLE_PRIVATE_KEY
    cert = GOOGLE_CERT

    user_claims = {
        "picture": "http://avatar",
        "name": "monalisa",
    }
    access_token = "b"
    half_hash = create_half_hash(access_token, "RS256")
    assert half_hash is not None
    payload = {
        "iss": "https://accounts.google.com",
        "azp": google_client_id,
        "aud": "1234987819200.apps.googleusercontent.com",
        "sub": login + "_id",
        "at_hash": half_hash.decode("ascii"),
        "hd": "example.com",
        "email": login,
        "email_verified": "true",
        "iat": int(time.time()),
        "exp": int(time.time()) + 100,
        "nonce": "0394852-3190485-2490358",
        **user_claims,
    }

    header = {"alg": "RS256", "kid": "1"}

    id_token = jwt.encode(header, payload, key).decode('ascii')
    response = {
        '/token': {'body': {'access_token': access_token, "id_token": id_token}},
        '/oauth2/v3/certs': {
            "body": {
                "keys": [
                    {
                        "kid": "1",
                        **JsonWebKey.import_key(cert, {"kty": "RSA"}).as_dict(),
                    }
                ]
            }
        },
        '/.well-known/openid-configuration': {
            "body": {
                "issuer": "https://accounts.google.com",
                "authorization_endpoint": "https://accounts.google.com/o/oauth2/v2/auth",  # noqa
                "device_authorization_endpoint": "https://oauth2.googleapis.com/device/code",  # noqa
                "token_endpoint": "https://oauth2.googleapis.com/token",
                "userinfo_endpoint": "https://openidconnect.googleapis.com/v1/userinfo",
                "jwks_uri": "https://www.googleapis.com/oauth2/v3/certs",
            }
        },
    }

    return "test_google", response


@pytest.fixture
def config_auth():
    return """
[github]
client_id = "aaa"
client_secret = "bbb"

[google]
client_id = "aaa"
client_secret = "bbb"

[jupyterhubauthenticator]
client_id = "test_client"
client_secret = "test_secret"
access_token_url = "http://jupyterhub/hub/api/oauth2/token"
authorize_url = "http://jupyterhub/hub/api/oauth2/authorize"
api_base_url = "http://jupyterhub/hub/api/"

[azuread]
client_id = "aaa"
client_secret = "bbb"
tenant_id = "3cd6563a-bbcf-4e30-bbc6-ea3445c75c79"
"""


@pytest.fixture(
    params=[
        "github_response",
        "google_response",
        "jupyter_response",
        "azuread_response",
    ]
)
def provider_spec(request):
    provider, response = request.getfixturevalue(request.param)
    return provider, response


@pytest.fixture
def oauth_server(request, config, app, provider_spec):
    provider, response = provider_spec

    server_app = AsyncPathMapDispatch(response)

    # we need to remove the client, because it might have been already
    # registered in quetz.main

    if provider == 'test_github':
        auth_module = auth_github.GithubAuthenticator
    elif provider == "test_google":
        auth_module = auth_google.GoogleAuthenticator
    elif provider == 'test_jupyter':
        auth_module = JupyterhubAuthenticator
    elif provider == 'test_azuread':
        auth_module = AzureADAuthenticator
    else:
        raise Exception(f"not recognised provider {provider}")

    module = auth_module(
        config,
        client_kwargs={'app': server_app},
        provider=provider,
        app=app,
    )

    yield module

    module.oauth._registry.pop(provider)
    module.oauth._clients.pop(provider)


@pytest.fixture
def routed_client(app, oauth_server):
    # need to prepend the routes so that frontend routes do not
    # take priority
    for route in oauth_server.router.routes:
        app.router.routes.insert(0, route)

    yield TestClient(app)

    for route in oauth_server.router.routes:
        app.router.routes.remove(route)


@pytest.mark.parametrize("config_extra", ["[users]\ncreate_default_channel = true"])
def test_config_create_default_channel(routed_client, db, oauth_server, config):
    response = routed_client.get(f'/auth/{oauth_server.provider}/authorize')

    assert response.status_code == 200

    user = db.query(User).filter(User.username == "user-with-role").one_or_none()
    assert user

    channel = db.query(Channel).filter(Channel.name == "user-with-role").one_or_none()

    assert channel
    assert user == channel.members[0].user


@pytest.mark.parametrize("default_role", [None, 'member'])
@pytest.mark.parametrize(
    "user_group,expected_role",
    [
        ("admins", "owner"),
        ("maintainers", "maintainer"),
        ("members", "member"),
        (None, None),
    ],
)
def test_config_create_user_with_role(
    routed_client, db, oauth_server, config, default_role, expected_role, login
):
    response = routed_client.get(f'/auth/{oauth_server.provider}/authorize')

    assert response.status_code == 200

    user = db.query(User).filter(User.username == login).one_or_none()
    assert user

    if expected_role:
        assert user.role == expected_role
    elif default_role:
        assert user.role == default_role
    else:
        assert user.role is None


@pytest.mark.parametrize("default_role", [None, 'member'])
@pytest.mark.parametrize("user_roles", [{"admins": ["other_provider:user-with-role"]}])
def test_config_create_user_with_role_in_different_provider(
    routed_client, db, oauth_server, config, default_role, login
):
    # if the user logins from a different provider than the one specified
    # in config, default_role should be assumed

    response = routed_client.get(f'/auth/{oauth_server.provider}/authorize')

    assert response.status_code == 200

    user = db.query(User).filter(User.username == login).one_or_none()
    assert user

    if default_role:
        assert user.role == default_role
    else:
        assert user.role is None


@pytest.fixture
def channel(db):
    channel = Channel(name="user-with-role", private=True)
    db.add(channel)
    db.commit()
    return channel


@pytest.mark.parametrize("config_extra", ["[users]\ncreate_default_channel = true"])
def test_config_create_default_channel_exists(routed_client, db, oauth_server, channel):
    response = routed_client.get(f'/auth/{oauth_server.provider}/authorize')

    assert response.status_code == 200

    user = db.query(User).filter(User.username == "user-with-role").one_or_none()
    assert user

    channel = db.query(Channel).filter(Channel.name == "user-with-role").one_or_none()

    assert channel
    assert user not in [member.user for member in channel.members]

    user_channel = (
        db.query(ChannelMember).filter(ChannelMember.user_id == user.id).one_or_none()
    )

    assert user_channel

    assert user_channel.channel_name.startswith("user-with-role")


@pytest.fixture
def user(dao: Dao, db):
    user = dao.create_user_with_role("existing_user", role=SERVER_OWNER)
    yield user
    db.delete(user)
    db.commit()


@pytest.mark.parametrize("default_role", ["member", "maintainer", "owner", None])
@pytest.mark.parametrize("login", ["existing_user", "new_user"])
def test_config_user_exists(
    routed_client, db, oauth_server, channel, user, login, default_role, app
):
    profile = Profile(user=user)
    identity = Identity(
        provider=oauth_server.provider, user=user, identity_id="existing_user_id"
    )
    db.add(identity)
    db.add(profile)
    db.commit()

    response = routed_client.get(f'/auth/{oauth_server.provider}/authorize')

    assert response.status_code == 200

    db.refresh(user)

    users = db.query(User).filter(User.username == login).all()
    assert len(users) == 1

    new_user = users[0]

    assert new_user
    assert new_user.username == login
    assert login != "existing_user" or new_user.role == SERVER_OWNER
    assert login == "existing_user" or new_user.role == default_role
    assert new_user.profile
    assert new_user.identities
    assert new_user.identities[0].provider == oauth_server.provider


@pytest.fixture
def some_identity(user, db):
    identity = Identity(
        provider="some-provider", user=user, identity_id='some-identity'
    )
    db.add(identity)
    db.commit()
    yield identity
    db.delete(identity)
    db.commit()


# db will automatically rollback, so need to disable running tests in nested transaction
@pytest.mark.parametrize("auto_rollback", [False])
@pytest.mark.parametrize("login", ["existing_user"])
def test_config_user_with_identity_exists(
    routed_client,
    db,
    oauth_server,
    user,
    login,
    some_identity,
):
    # we should not be allowed to associate a social account with a user that
    # already has an identity from a different provider

    response = routed_client.get(f'/auth/{oauth_server.provider}/authorize')
    db.refresh(user)
    assert len(user.identities) == 1
    assert user.identities[0].provider == "some-provider"
    assert response.status_code == 422



def can_channel_synchronize(channel):
    return channel.mirror_channel_url and (channel.mirror_mode == "mirror")


def can_channel_synchronize_metrics(channel):
    return not channel.mirror_channel_url


def can_channel_generate_indexes(channel):
    return True


def can_channel_validate_package_cache(channel):
    return True


def can_channel_reindex(channel):
    return True


def can_cleanup(channel):
    return True



from typing import Callable, Optional, Union

import requests

from quetz.config import Config
from quetz.dao import Dao
from quetz.tasks.workers import job_wrapper


class MockWorker:
    "synchronous worker for testing"

    def __init__(
        self,
        config: Config,
        db,
        dao: Dao,
        session: Optional[requests.Session] = None,
    ):
        self.db = db
        self.dao = dao
        self.session = session
        self.config = config

    def execute(self, func: Union[Callable, bytes], *args, **kwargs):
        resources = {
            "db": self.db,
            "dao": self.dao,
            "pkgstore": self.config.get_package_store(),
        }

        if self.session:
            resources['session'] = self.session

        kwargs.update(resources)
        job_wrapper(func, self.config, *args, **kwargs)



"""Update scoped API key uploader id

Revision ID: 3ba25f23fb7d
Revises: d212023a8e0b
Create Date: 2023-08-02 08:03:09.961559

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = '3ba25f23fb7d'
down_revision = 'd212023a8e0b'
branch_labels = None
depends_on = None


def upgrade():
    package_versions = sa.sql.table("package_versions", sa.sql.column("uploader_id"))
    conn = op.get_bind()
    # Get all user_id/owner_id from channel scoped API keys
    # (user is anonymous - username is null)
    res = conn.execute(
        sa.text(
            """SELECT api_keys.user_id, api_keys.owner_id FROM api_keys
            INNER JOIN users ON users.id = api_keys.user_id
            WHERE users.username is NULL;
            """
        )
    )
    results = res.fetchall()
    # Replace the uploader with the key owner (real user instead of the anonymous one)
    for result in results:
        op.execute(
            package_versions.update()
            .where(package_versions.c.uploader_id == result[0])
            .values(uploader_id=result[1])
        )


def downgrade():
    package_versions = sa.sql.table("package_versions", sa.sql.column("uploader_id"))
    conn = op.get_bind()
    # Get all user_id/owner_id from channel scoped API keys
    # (user is anonymous - username is null)
    res = conn.execute(
        sa.text(
            """SELECT api_keys.user_id, api_keys.owner_id FROM api_keys
            INNER JOIN users ON users.id = api_keys.user_id
            WHERE users.username is NULL;
            """
        )
    )
    results = res.fetchall()
    # Replace the uploader with the key anonymous user
    for result in results:
        op.execute(
            package_versions.update()
            .where(package_versions.c.uploader_id == result[1])
            .values(uploader_id=result[0])
        )



# Copyright 2020 QuantStack
# Distributed under the terms of the Modified BSD License.

import logging
import logging.config
import os
from distutils.util import strtobool
from secrets import token_bytes
from typing import Any, Dict, Iterable, List, NamedTuple, Optional, Type, Union

import appdirs
import pluggy
import toml

from quetz import hooks, pkgstores
from quetz.errors import ConfigError

_filename = "config.toml"
_env_prefix = "QUETZ_"
_env_config_file = "CONFIG_FILE"
_site_dir = appdirs.site_config_dir("quetz")
_user_dir = appdirs.user_config_dir("quetz")

PAGINATION_LIMIT = 20


class ConfigEntry(NamedTuple):
    name: str
    cast: Type
    default: Any = None
    required: bool = True

    def full_name(self, section=""):
        if section:
            section += "_"
        return f"{section}{self.name}"

    def env_var(self, section=""):
        return f"{_env_prefix}{self.full_name(section).upper()}"

    def casted(self, value):
        if self.cast is bool:
            try:
                value = strtobool(str(value))
            except ValueError as e:
                raise ConfigError(f"{self.name}: {e}")

        return self.cast(value)


class ConfigSection(NamedTuple):
    name: str
    entries: List[ConfigEntry]
    required: bool = True


class Config:
    _config_map = [
        ConfigSection(
            "general",
            [
                ConfigEntry("package_unpack_threads", int, 1),
                ConfigEntry("frontend_dir", str, default=""),
                ConfigEntry("redirect_http_to_https", bool, False),
            ],
        ),
        ConfigSection(
            "cors",
            [
                ConfigEntry("allow_origins", list, []),
                ConfigEntry("allow_credentials", bool, True),
                ConfigEntry("allow_methods", bool, ["*"]),
                ConfigEntry("allow_headers", bool, ["*"]),
            ],
            required=False,
        ),
        ConfigSection(
            "github",
            [
                ConfigEntry("client_id", str),
                ConfigEntry("client_secret", str),
            ],
            required=False,
        ),
        ConfigSection(
            "gitlab",
            [
                ConfigEntry("url", str, default="https://gitlab.com"),
                ConfigEntry("client_id", str),
                ConfigEntry("client_secret", str),
            ],
            required=False,
        ),
        ConfigSection(
            "azuread",
            [
                ConfigEntry("client_id", str),
                ConfigEntry("client_secret", str),
                ConfigEntry("tenant_id", str),
            ],
            required=False,
        ),
        ConfigSection(
            "sqlalchemy",
            [
                ConfigEntry("database_url", str),
                ConfigEntry("database_plugin_path", str, default="", required=False),
                ConfigEntry("echo_sql", bool, default=False, required=False),
                ConfigEntry("postgres_pool_size", int, default=10, required=False),
                ConfigEntry("postgres_max_overflow", int, default=100, required=False),
            ],
        ),
        ConfigSection(
            "session",
            [ConfigEntry("secret", str), ConfigEntry("https_only", bool, default=True)],
        ),
        ConfigSection(
            "local_store",
            [
                ConfigEntry("redirect_enabled", bool, default=False),
                ConfigEntry("redirect_endpoint", str, default="/files"),
                ConfigEntry("redirect_secret", str, default=""),
                ConfigEntry("redirect_expiration", int, default="3600"),
            ],
        ),
        ConfigSection(
            "s3",
            [
                ConfigEntry("access_key", str, default=""),
                ConfigEntry("secret_key", str, default=""),
                ConfigEntry("url", str, default=""),
                ConfigEntry("region", str, default=""),
                ConfigEntry("bucket_prefix", str, default=""),
                ConfigEntry("bucket_suffix", str, default=""),
            ],
            required=False,
        ),
        ConfigSection(
            "azure_blob",
            [
                ConfigEntry("account_name", str, default=""),
                ConfigEntry("account_access_key", str, default=""),
                ConfigEntry("conn_str", str, default=""),
                ConfigEntry("container_prefix", str, default=""),
                ConfigEntry("container_suffix", str, default=""),
            ],
            required=False,
        ),
        ConfigSection(
            "gcs",
            [
                ConfigEntry("project", str, default=""),
                ConfigEntry("token", str, default=""),
                ConfigEntry("bucket_prefix", str, default=""),
                ConfigEntry("bucket_suffix", str, default=""),
                ConfigEntry("cache_timeout", int, default=None),
                ConfigEntry("region", str, default=None),
            ],
            required=False,
        ),
        ConfigSection(
            "google",
            [ConfigEntry("client_id", str), ConfigEntry("client_secret", str)],
            required=False,
        ),
        ConfigSection(
            "logging",
            [
                ConfigEntry("level", str, default="INFO"),
                ConfigEntry("file", str, default=""),
            ],
            required=False,
        ),
        ConfigSection(
            "users",
            [
                ConfigEntry("admins", list, default=list),
                ConfigEntry("maintainers", list, default=list),
                ConfigEntry("members", list, default=list),
                ConfigEntry("default_role", str, required=False),
                ConfigEntry("collect_emails", bool, default=False, required=False),
                ConfigEntry(
                    "create_default_channel", bool, default=False, required=False
                ),
            ],
            required=False,
        ),
        ConfigSection(
            "worker",
            [
                ConfigEntry("type", str, default="thread"),
                ConfigEntry("redis_ip", str, default="127.0.0.1"),
                ConfigEntry("redis_port", int, default=6379),
                ConfigEntry("redis_db", int, default=0),
            ],
            required=False,
        ),
        ConfigSection(
            "plugins",
            [
                ConfigEntry("enabled", list, default=list),
            ],
        ),
        ConfigSection(
            "mirroring",
            [
                ConfigEntry("batch_length", int, default=10),
                ConfigEntry("batch_size", int, default=int(1e8)),
                ConfigEntry("num_parallel_downloads", int, default=int(10)),
            ],
        ),
        ConfigSection(
            "quotas",
            [
                ConfigEntry("channel_quota", int, required=False),
            ],
            required=False,
        ),
        ConfigSection(
            "profiling",
            [
                ConfigEntry("enable_sampling", bool, required=False, default=False),
                ConfigEntry("interval_seconds", float, required=False, default=0.001),
            ],
            required=False,
        ),
    ]
    _config_dirs = [_site_dir, _user_dir]
    _config_files = [os.path.join(d, _filename) for d in _config_dirs]

    _instances: Dict[Optional[str], "Config"] = {}

    def __new__(cls, deployment_config: str = None):
        if not deployment_config and None in cls._instances:
            return cls._instances[None]

        try:
            path = os.path.abspath(cls.find_file(deployment_config))
        except TypeError:
            # if not config path exists, set it to empty string.
            path = ""

        if path not in cls._instances:
            config = super().__new__(cls)
            config.init(path)
            cls._instances[path] = config
            # optimization - for default config path we also store the instance
            # under None key
            if not deployment_config:
                cls._instances[None] = config
        return cls._instances[path]

    def __getattr__(self, name: str) -> Any:
        super().__getattr__(self, name)

    @classmethod
    def find_file(cls, deployment_config: str = None):
        config_file_env = os.getenv(f"{_env_prefix}{_env_config_file}")
        deployment_config_files = []
        for f in (deployment_config, config_file_env):
            if f and os.path.isfile(f):
                deployment_config_files.append(f)

        # In order, get configuration from:
        # _site_dir, _user_dir, deployment_config, config_file_env
        for f in cls._config_files + deployment_config_files:
            if os.path.isfile(f):
                return f

    def init(self, path: str) -> None:
        """Load configurations from various places.

        Order of importance for configuration is:
        host < user profile < deployment < configuration file from env var < value from
        env var

        Parameters
        ----------
        deployment_config : str, optional
            The configuration stored at deployment level
        """

        self.config: Dict[str, Any] = {}

        # only try to get config from config file if it exists.
        if path:
            self.config.update(self._read_config(path))

        self.config.update(self._get_environ_config())
        self._trigger_update_config()

    def _trigger_update_config(self):
        def set_entry_attr(entry, section=""):
            value = self._get_value(entry, section)

            setattr(self, entry.full_name(section), value)

        for item in self._config_map:
            if isinstance(item, ConfigSection) and (
                item.required or item.name in self.config
            ):
                for entry in item.entries:
                    set_entry_attr(entry, item.name)
            elif isinstance(item, ConfigEntry):
                set_entry_attr(item)

    def _get_value(
        self, entry: ConfigEntry, section: str = ""
    ) -> Union[str, bool, None]:
        """Get an entry value from a configuration mapping.

        Parameters
        ----------
        entry : ConfigEntry
            The entry to search
        section : str
            The section the entry belongs to

        Returns
        -------
        value : Union[str, bool]
            The entry value
        """
        try:
            if section:
                value = self.config[section][entry.name]
            else:
                value = self.config[entry.name]

            return entry.casted(value)

        except KeyError:
            if entry.default is not None:
                if callable(entry.default):
                    return entry.default()
                return entry.default

        msg = f"'{entry.name}' unset but no default specified"
        if section:
            msg += f" for section '{section}'"

        if entry.required:
            raise ConfigError(msg)

        return None

    def _read_config(self, filename: str) -> Dict[str, Any]:
        """Read a configuration file from its path.

        Parameters
        ----------
        filename : str
            The path of the configuration file

        Returns
        -------
        configuration : Dict[str, str]
            The mapping of configuration variables found in the file
        """
        with open(filename) as f:
            try:
                return dict(toml.load(f))
            except toml.TomlDecodeError as e:
                raise ConfigError(f"failed to load config file '{filename}': {e}")

    def _find_first_level_config(
        self, section_name: str
    ) -> Union[ConfigSection, ConfigEntry, None]:
        """Find the section or entry at first level of config_map.

        Parameters
        ----------
        section_name : str
            The name of the section to find.

        Returns
        -------
        section : Union[ConfigSection, ConfigEntry, None]
            The section or entry found, else None.
        """
        for item in self._config_map:
            if section_name == item.name:
                return item
        return None

    def _get_environ_config(self) -> Dict[str, Any]:
        """Looks into environment variables if some matches with config_map.

        Returns
        -------
        configuration : Dict[str, str]
            The mapping of configuration variables found in environment variables.
        """
        config: Dict[str, Any] = {}

        # get QUETZ environment variables.
        quetz_var = {
            key: value
            for key, value in os.environ.items()
            if key.startswith(_env_prefix)
        }
        for var, value in quetz_var.items():
            splitted_key = var.split('_')
            config_key = splitted_key[1].lower()
            idx = 2

            # look for the first level of config_map.
            # It must be done in loop as the key itself can contains '_'.
            first_level = None
            while idx < len(splitted_key):
                first_level = self._find_first_level_config(config_key)
                if first_level:
                    break
                config_key += f"_{ splitted_key[idx].lower()}"
                idx += 1

            # no first_level found, the variable is useless.
            if not first_level:
                continue
            # the first level is an entry, add it to the config.
            if isinstance(first_level, ConfigEntry):
                config[first_level.name] = value
            # the first level is a section.
            elif isinstance(first_level, ConfigSection):
                entry = "_".join(splitted_key[idx:]).lower()
                # the entry does not exist in section, the variable is useless.
                if entry not in [
                    section_entry.name for section_entry in first_level.entries
                ]:
                    continue
                # add the entry to the config.
                if first_level.name not in config:
                    config[first_level.name]: Dict[str, Any] = {}
                config[first_level.name]["_".join(splitted_key[idx:]).lower()] = value

        return config

    def get_package_store(self) -> pkgstores.PackageStore:
        """Return the appropriate package store as set in the config.

        Returns
        -------
        package_store : pkgstores.PackageStore
            The package store instance to enact package operations against
        """
        if self.config.get('s3'):
            return pkgstores.S3Store(
                {
                    'key': self.s3_access_key,
                    'secret': self.s3_secret_key,
                    'url': self.s3_url,
                    'region': self.s3_region,
                    'bucket_prefix': self.s3_bucket_prefix,
                    'bucket_suffix': self.s3_bucket_suffix,
                }
            )
        elif self.config.get('azure_blob'):
            return pkgstores.AzureBlobStore(
                {
                    'account_name': self.azure_blob_account_name,
                    'account_access_key': self.azure_blob_account_access_key,
                    'conn_str': self.azure_blob_conn_str,
                    'container_prefix': self.azure_blob_container_prefix,
                    'container_suffix': self.azure_blob_container_suffix,
                }
            )
        elif self.config.get('gcs'):
            return pkgstores.GoogleCloudStorageStore(
                {
                    'project': self.gcs_project,
                    'token': self.gcs_token,
                    'bucket_prefix': self.gcs_bucket_prefix,
                    'bucket_suffix': self.gcs_bucket_suffix,
                    'cache_timeout': self.gcs_cache_timeout,
                    'region': self.gcs_region,
                }
            )
        else:
            return pkgstores.LocalStore(
                {
                    'channels_dir': 'channels',
                    'redirect_enabled': self.local_store_redirect_enabled,
                    'redirect_endpoint': self.local_store_redirect_endpoint,
                    'redirect_secret': self.local_store_redirect_secret,
                    'redirect_expiration': int(self.local_store_redirect_expiration),
                }
            )

    def configured_section(self, section: str) -> bool:
        """Return if a given section has been configured.

        Parameters
        ----------
        provider: str
            The section name in config

        Returns
        -------
        bool
            Wether or not the given section is configured
        """

        return bool(self.config.get(section))

    def register(self, extra_config: Iterable[ConfigSection]):
        """Register additional config variables"""
        self._config_map += extra_config
        self._trigger_update_config()


def create_config(
    client_id: str = "",
    client_secret: str = "",
    database_url: str = "sqlite:///./quetz.sqlite",
    secret: str = token_bytes(32).hex(),
    https: str = 'true',
) -> str:
    """Create a configuration file from a template.

    Parameters
    ----------
    client_id : str, optional
        The client ID {default=""}
    client_secret : str, optional
        The client secret {default=""}
    database_url : str, optional
        The URL of the database {default="sqlite:///./quetz.sqlite"}
    secret : str, optional
        The secret of the session {default=randomly create}
    https : str, optional
        Whether to use HTTPS, or not {default="true"}

    Returns
    -------
    configuration : str
        The configuration
    """
    with open(os.path.join(os.path.dirname(__file__), _filename), 'r') as f:
        config = ''.join(f.readlines())

    return config.format(client_id, client_secret, database_url, secret, https)


def colourized_formatter(fmt="", use_colors=True):
    try:
        from uvicorn.logging import ColourizedFormatter

        return ColourizedFormatter(fmt, use_colors=use_colors)
    except ImportError:
        return logging.Formatter(fmt)


def get_logger_config(config, loggers):
    if hasattr(config, "logging_level"):
        log_level = config.logging_level
    else:
        log_level = "INFO"

    if hasattr(config, "logging_file"):
        filename = config.logging_file
    else:
        filename = None

    log_level = os.environ.get("QUETZ_LOG_LEVEL", log_level)

    log_level = log_level.upper()

    handlers = ["console"]
    if filename:
        handlers.append("file")

    LOG_FORMATTERS = {
        "colour": {
            "()": "quetz.config.colourized_formatter",
            "fmt": "%(levelprefix)s [%(name)s] %(message)s",
            "use_colors": True,
        },
        "basic": {"format": "%(levelprefix)s [%(name)s] %(message)s"},
        "timestamp": {"format": '%(asctime)s %(levelname)s %(name)s  %(message)s'},
    }

    curdir = os.getcwd()

    LOG_HANDLERS = {
        "console": {
            "class": "logging.StreamHandler",
            "formatter": "colour",
            "level": log_level,
            "stream": "ext://sys.stderr",
        },
        "file": {
            "class": "logging.FileHandler",
            "formatter": "timestamp",
            "filename": filename or os.path.join(curdir, "quetz.log"),
            "level": log_level,
        },
    }

    LOGGERS = {k: {"level": log_level, "handlers": handlers} for k in loggers}

    LOG_CONFIG = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": LOG_FORMATTERS,
        "handlers": LOG_HANDLERS,
        "loggers": LOGGERS,
    }

    return LOG_CONFIG


def configure_logger(config=None, loggers=("quetz", "urllib3.util.retry", "alembic")):
    """Get quetz logger"""

    log_config = get_logger_config(config, loggers)

    logging.config.dictConfig(log_config)


def get_plugin_manager(config=None) -> pluggy.PluginManager:
    """Create an instance of plugin manager."""

    if not config:
        config = Config()

    pm = pluggy.PluginManager("quetz")
    pm.add_hookspecs(hooks)
    if config.configured_section("plugins"):
        for name in config.plugins_enabled:
            pm.load_setuptools_entrypoints("quetz", name)
    else:
        pm.load_setuptools_entrypoints("quetz")
    return pm



"""create tables

Revision ID: a80fb051a659
Revises:
Create Date: 2020-11-24 12:05:33.044466

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = 'a80fb051a659'
down_revision = None
branch_labels = ("quetz",)
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        'channels',
        sa.Column('name', sa.String(), nullable=False),
        sa.Column('description', sa.String(), nullable=True),
        sa.Column('private', sa.Boolean(), nullable=True),
        sa.Column('mirror_channel_url', sa.String(), nullable=True),
        sa.Column('mirror_mode', sa.String(), nullable=True),
        sa.Column('timestamp_mirror_sync', sa.Integer(), nullable=True),
        sa.PrimaryKeyConstraint('name'),
    )
    op.create_index(op.f('ix_channels_name'), 'channels', ['name'], unique=False)
    op.create_table(
        'users',
        sa.Column('id', sa.LargeBinary(length=16), nullable=False),
        sa.Column('username', sa.String(), nullable=True),
        sa.Column('role', sa.String(), nullable=True),
        sa.PrimaryKeyConstraint('id'),
    )
    op.create_index(op.f('ix_users_id'), 'users', ['id'], unique=False)
    op.create_index(op.f('ix_users_username'), 'users', ['username'], unique=True)
    op.create_table(
        'api_keys',
        sa.Column('key', sa.String(), nullable=False),
        sa.Column('description', sa.String(), nullable=True),
        sa.Column('deleted', sa.Boolean(), nullable=True),
        sa.Column('user_id', sa.LargeBinary(length=16), nullable=True),
        sa.Column('owner_id', sa.LargeBinary(length=16), nullable=True),
        sa.ForeignKeyConstraint(
            ['owner_id'],
            ['users.id'],
        ),
        sa.ForeignKeyConstraint(
            ['user_id'],
            ['users.id'],
        ),
        sa.PrimaryKeyConstraint('key'),
    )
    op.create_index(op.f('ix_api_keys_key'), 'api_keys', ['key'], unique=False)
    op.create_table(
        'channel_members',
        sa.Column('channel_name', sa.String(), nullable=False),
        sa.Column('user_id', sa.LargeBinary(length=16), nullable=False),
        sa.Column('role', sa.String(), nullable=True),
        sa.ForeignKeyConstraint(
            ['channel_name'],
            ['channels.name'],
        ),
        sa.ForeignKeyConstraint(
            ['user_id'],
            ['users.id'],
        ),
        sa.PrimaryKeyConstraint('channel_name', 'user_id'),
    )
    op.create_index(
        op.f('ix_channel_members_channel_name'),
        'channel_members',
        ['channel_name'],
        unique=False,
    )
    op.create_index(
        op.f('ix_channel_members_user_id'), 'channel_members', ['user_id'], unique=False
    )
    op.create_table(
        'identities',
        sa.Column('provider', sa.String(), nullable=False),
        sa.Column('identity_id', sa.String(), nullable=False),
        sa.Column('username', sa.String(), nullable=True),
        sa.Column('user_id', sa.LargeBinary(length=16), nullable=True),
        sa.ForeignKeyConstraint(
            ['user_id'],
            ['users.id'],
        ),
        sa.PrimaryKeyConstraint('provider', 'identity_id'),
    )
    op.create_index(
        'identity_index', 'identities', ['provider', 'identity_id'], unique=True
    )
    op.create_table(
        'packages',
        sa.Column('name', sa.String(), nullable=False),
        sa.Column('channel_name', sa.String(), nullable=False),
        sa.Column('description', sa.Text(), nullable=True),
        sa.Column('summary', sa.Text(), nullable=True),
        sa.Column('channeldata', sa.String(), nullable=True),
        sa.ForeignKeyConstraint(
            ['channel_name'],
            ['channels.name'],
        ),
        sa.PrimaryKeyConstraint('name', 'channel_name'),
    )
    op.create_index(
        op.f('ix_packages_channel_name'), 'packages', ['channel_name'], unique=False
    )
    op.create_index(op.f('ix_packages_name'), 'packages', ['name'], unique=False)
    op.create_table(
        'profiles',
        sa.Column('name', sa.String(), nullable=True),
        sa.Column('avatar_url', sa.String(), nullable=True),
        sa.Column('user_id', sa.LargeBinary(length=16), nullable=False),
        sa.ForeignKeyConstraint(
            ['user_id'],
            ['users.id'],
        ),
        sa.PrimaryKeyConstraint('user_id'),
    )
    op.create_table(
        'package_members',
        sa.Column('channel_name', sa.String(), nullable=False),
        sa.Column('package_name', sa.String(), nullable=False),
        sa.Column('user_id', sa.LargeBinary(length=16), nullable=False),
        sa.Column('role', sa.String(), nullable=True),
        sa.ForeignKeyConstraint(
            ['channel_name', 'package_name'],
            ['packages.channel_name', 'packages.name'],
        ),
        sa.ForeignKeyConstraint(
            ['channel_name'],
            ['channels.name'],
        ),
        sa.ForeignKeyConstraint(
            ['user_id'],
            ['users.id'],
        ),
        sa.PrimaryKeyConstraint('channel_name', 'package_name', 'user_id'),
    )
    op.create_index(
        op.f('ix_package_members_channel_name'),
        'package_members',
        ['channel_name'],
        unique=False,
    )
    op.create_index(
        op.f('ix_package_members_package_name'),
        'package_members',
        ['package_name'],
        unique=False,
    )
    op.create_index(
        op.f('ix_package_members_user_id'), 'package_members', ['user_id'], unique=False
    )
    op.create_table(
        'package_versions',
        sa.Column('id', sa.LargeBinary(length=16), nullable=False),
        sa.Column('channel_name', sa.String(), nullable=True),
        sa.Column('package_name', sa.String(), nullable=True),
        sa.Column(
            'package_format',
            sa.Enum('tarbz2', 'conda', name='packageformatenum'),
            nullable=True,
        ),
        sa.Column('platform', sa.String(), nullable=True),
        sa.Column('version', sa.String(), nullable=True),
        sa.Column('build_string', sa.String(), nullable=True),
        sa.Column('build_number', sa.Integer(), nullable=True),
        sa.Column('version_order', sa.Integer(), nullable=True),
        sa.Column('filename', sa.String(), nullable=True),
        sa.Column('info', sa.String(), nullable=True),
        sa.Column('uploader_id', sa.LargeBinary(length=16), nullable=True),
        sa.Column(
            'time_created',
            sa.DateTime(timezone=True),
            server_default=sa.text('(CURRENT_TIMESTAMP)'),
            nullable=True,
        ),
        sa.Column(
            'time_modified',
            sa.DateTime(timezone=True),
            server_default=sa.text('(CURRENT_TIMESTAMP)'),
            nullable=True,
        ),
        sa.ForeignKeyConstraint(
            ['channel_name', 'package_name'],
            ['packages.channel_name', 'packages.name'],
        ),
        sa.ForeignKeyConstraint(
            ['channel_name'],
            ['channels.name'],
        ),
        sa.ForeignKeyConstraint(
            ['uploader_id'],
            ['users.id'],
        ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint(
            'channel_name',
            'package_name',
            'package_format',
            'platform',
            'version',
            'build_string',
            'build_number',
            name='package_version_index',
        ),
    )
    op.create_index(
        'package_version_name_index',
        'package_versions',
        ['channel_name', 'package_name'],
        unique=False,
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index('package_version_name_index', table_name='package_versions')
    op.drop_table('package_versions')
    op.drop_index(op.f('ix_package_members_user_id'), table_name='package_members')
    op.drop_index(op.f('ix_package_members_package_name'), table_name='package_members')
    op.drop_index(op.f('ix_package_members_channel_name'), table_name='package_members')
    op.drop_table('package_members')
    op.drop_table('profiles')
    op.drop_index(op.f('ix_packages_name'), table_name='packages')
    op.drop_index(op.f('ix_packages_channel_name'), table_name='packages')
    op.drop_table('packages')
    op.drop_index('identity_index', table_name='identities')
    op.drop_table('identities')
    op.drop_index(op.f('ix_channel_members_user_id'), table_name='channel_members')
    op.drop_index(op.f('ix_channel_members_channel_name'), table_name='channel_members')
    op.drop_table('channel_members')
    op.drop_index(op.f('ix_api_keys_key'), table_name='api_keys')
    op.drop_table('api_keys')
    op.drop_index(op.f('ix_users_username'), table_name='users')
    op.drop_index(op.f('ix_users_id'), table_name='users')
    op.drop_table('users')
    op.drop_index(op.f('ix_channels_name'), table_name='channels')
    op.drop_table('channels')
    # ### end Alembic commands ###



# Copyright 2020 Codethink Ltd
# Distributed under the terms of the Modified BSD License.


class QuetzError(Exception):
    pass


class DBError(QuetzError):
    pass


class ConfigError(QuetzError):
    pass


class ValidationError(QuetzError):
    pass


class QuotaError(ValidationError):
    pass


class TaskError(QuetzError):
    pass



# Copyright 2020 Codethink Ltd
# Distributed under the terms of the Modified BSD License.

import json
import logging
import numbers
import os
import tempfile
import uuid
from datetime import datetime, timezone
from pathlib import Path

from jinja2 import Environment, PackageLoader, select_autoescape
from jinja2.exceptions import UndefinedError

import quetz.config
from quetz import channel_data, repo_data
from quetz.condainfo import MAX_CONDA_TIMESTAMP
from quetz.db_models import PackageVersion
from quetz.utils import add_static_file, add_temp_static_file

_iec_prefixes = (
    # IEEE 1541 - IEEE Standard for Prefixes for Binary Multiples
    # ISO/IEC 80000-13:2008 Clause 4 binary prefixes
    # https://en.wikipedia.org/wiki/Binary_prefix
    (1024 * 1024 * 1024 * 1024, "{:.2f} TiB"),
    (1024 * 1024 * 1024, "{:.2f} GiB"),
    (1024 * 1024, "{:.1f} MiB"),
    (1024, "{:.0f} KiB"),
    (1, "{:.0f} B"),
)

logger = logging.getLogger("quetz")


def _iec_bytes(n):
    # Return human-readable string representing n in bytes in IEC format
    for e, f in _iec_prefixes:
        try:
            if n >= e:
                return f.format(n / e)
        except UndefinedError:
            logger.debug("Package size is undefined.")
    return f"{n} B"


def _strftime(date, date_format):
    if isinstance(date, numbers.Real):
        if date > MAX_CONDA_TIMESTAMP:
            date //= 1000
        date = datetime.fromtimestamp(date, timezone.utc)

    if isinstance(date, datetime):
        return date.strftime(date_format)


def _opt_href(text, link):
    if link:
        return f"<a href={link}>{text}</a>"
    else:
        return text


def _jinjaenv():
    env = Environment(
        loader=PackageLoader("quetz", "templates"),
        autoescape=select_autoescape(["html", "xml"]),
    )
    env.trim_blocks = True
    env.lstrip_blocks = True
    env.filters["iec_bytes"] = _iec_bytes
    env.filters["strftime"] = _strftime
    env.filters["opt_href"] = _opt_href
    return env


_subdir_order = {
    # This determines the ordering of subdirectories in index.html
    "linux-64": "!000",
    "osx-64": "!001",
    "win-64": "!002",
    # New architectures go here
    "noarch": "~~~~~~~",  # sorts last
}


def _subdir_key(dir):
    return _subdir_order.get(dir, dir)


def validate_packages(dao, pkgstore, channel_name):
    # for now we're just validating the size of the uploaded file
    logger.info("Starting package validation")

    if type(pkgstore).__name__ == "S3Store":
        fs_chan = pkgstore._bucket_map(channel_name)
    elif type(pkgstore).__name__ == "LocalStore":
        fs_chan = os.path.abspath(os.path.join(pkgstore.channels_dir, channel_name))
    else:
        raise KeyError(type(pkgstore).__name__)

    logger.info(f"Checking {fs_chan}")
    ls_dirs = pkgstore.fs.ls(f"{fs_chan}", detail=True)
    dirs = [d['name'].rsplit('/', 1)[1] for d in ls_dirs if d['type'] == 'directory']

    for subdir in dirs:
        ls_result = pkgstore.fs.ls(f"{fs_chan}/{subdir}", detail=True)

        ls_result_set = set([(res["name"].rsplit('/', 1)[1]) for res in ls_result])
        db_result = [
            (res.filename, json.loads(res.info)["size"])
            for res in dao.db.query(PackageVersion).filter(
                PackageVersion.channel_name == channel_name,
                PackageVersion.platform == subdir,
            )
        ]
        db_result_set = set((res[0] for res in db_result))

        difference = ls_result_set ^ db_result_set

        logger.info(f"Differing files: {difference}")

        in_ls_not_db = ls_result_set - db_result_set
        in_db_not_ls = db_result_set - ls_result_set

        valid, inexistant, wrong_size = 0, 0, 0

        # remove all files that are in database and not uploaded
        for f in in_db_not_ls:
            logger.warning(f"Removing non-existent file from database {f}")
            db_pkg_to_delete = (
                dao.db.query(PackageVersion)
                .filter(
                    PackageVersion.channel_name == channel_name,
                    PackageVersion.platform == subdir,
                    PackageVersion.filename == f,
                )
                .one()
            )
            dao.db.delete(db_pkg_to_delete)
            inexistant += 1

        db_dict = dict(db_result)
        for f in ls_result:
            filename = f["name"].rsplit('/', 1)[1]
            if filename in db_dict:
                if db_dict[filename] != f["size"]:
                    # size of file in db and on filesystem does not match!
                    logger.error(
                        f"File size differs for {filename}: "
                        f"{f['size']} vs {db_dict[filename]}"
                    )
                    pkgstore.delete_file(channel_name, f"{subdir}/{filename}")
                    db_pkg_to_delete = (
                        dao.db.query(PackageVersion)
                        .filter(
                            PackageVersion.channel_name == channel_name,
                            PackageVersion.platform == subdir,
                            PackageVersion.filename == filename,
                        )
                        .one()
                    )
                    dao.db.delete(db_pkg_to_delete)
                    wrong_size += 1
                else:
                    valid += 1

        dao.db.commit()

        logger.info(f"\n\nSUBDIR: {subdir}")
        logger.info(f"On filesystem: {in_ls_not_db}")
        logger.info(f"In database, not uploaded: {in_db_not_ls}")

        logger.info(f"Valid files: {valid}")
        logger.info(f"Wrong size: {wrong_size}")
        logger.info(f"Not uploaded: {inexistant}")

    update_indexes(dao, pkgstore, channel_name)


def update_indexes(dao, pkgstore, channel_name, subdirs=None):
    jinjaenv = _jinjaenv()
    channeldata = channel_data.export(dao, channel_name)

    if subdirs is None:
        subdirs = sorted(channeldata["subdirs"], key=_subdir_key)

    # Generate channeldata.json and its compressed version
    chandata_json = json.dumps(channeldata, indent=2, sort_keys=False)
    add_static_file(chandata_json, channel_name, None, "channeldata.json", pkgstore)

    # Generate index.html for the "root" directory
    channel_index = jinjaenv.get_template("channeldata-index.html.j2").render(
        title=channel_name,
        packages=channeldata["packages"],
        subdirs=subdirs,
        current_time=datetime.now(timezone.utc),
    )

    add_static_file(channel_index, channel_name, None, "index.html", pkgstore)

    # NB. No rss.xml is being generated here
    files = {}
    packages = {}
    subdir_template = jinjaenv.get_template("subdir-index.html.j2")

    tempdir = tempfile.TemporaryDirectory()
    tempdir_path = Path(tempdir.name)

    pm = quetz.config.get_plugin_manager()

    for sdir in subdirs:
        logger.debug(f"creating indexes for subdir {sdir} of channel {channel_name}")
        raw_repodata = repo_data.export(dao, channel_name, sdir)
        try:
            logger.debug(f"Starting post_index_creation for {sdir} of {channel_name}")
            pm.hook.post_index_creation(
                raw_repodata=raw_repodata, channel_name=channel_name, subdir=sdir
            )
            logger.debug(f"Finished post_index_creation for {sdir} of {channel_name}")
        except Exception:
            logger.exception("Exception post_index_creation:")

        files[sdir] = []
        packages[sdir] = raw_repodata["packages"] | raw_repodata["packages.conda"]

        repodata = json.dumps(raw_repodata, indent=2, sort_keys=False)

        add_temp_static_file(
            repodata, channel_name, sdir, "repodata.json", tempdir_path, files
        )

    try:
        logger.debug(f"Starting post_package_indexing for {channel_name}")
        pm.hook.post_package_indexing(
            tempdir=tempdir_path,
            channel_name=channel_name,
            subdirs=subdirs,
            files=files,
            packages=packages,
        )
        logger.debug(f"Finished post_package_indexing for {channel_name}")
    except Exception:
        logger.exception("Exception post_package_indexing:")

    for sdir in subdirs:
        # Generate subdir index.html
        subdir_index_html = subdir_template.render(
            title=f"{channel_name}/{sdir}",
            packages=packages[sdir],
            current_time=datetime.now(timezone.utc),
            add_files=files[sdir],
        )
        add_static_file(subdir_index_html, channel_name, sdir, "index.html", pkgstore)

    # recursively walk through the tree
    tmp_suffix = uuid.uuid4().hex
    after_upload_move = []
    for path in tempdir_path.rglob('*.*'):
        # check whether the path is an actual file.
        # this fixes https://github.com/mamba-org/quetz/issues/540
        if not path.is_file():
            continue
        rel_path = path.relative_to(tempdir_path)
        to_upload = open(path, 'rb')
        if len(rel_path.parts) == 2:
            channel_name, filename = rel_path.parts
            dest = f"{filename}{tmp_suffix}"
        elif len(rel_path.parts) == 3:
            channel_name, sdir, filename = rel_path.parts
            dest = f"{sdir}/{filename}{tmp_suffix}"
        else:
            raise NotImplementedError(
                "We can only handle channel_name/subdir/file.xyz OR"
                "channel_name/file.xyz"
            )

        logger.debug(f"Uploading {path} -> {channel_name} {dest}")
        if type(pkgstore).__name__ == "S3Store":
            with pkgstore._get_fs() as fs:
                bucket = pkgstore._bucket_map(channel_name)
                fs.put_file(path, f"{bucket}/{dest}")
        else:
            pkgstore.add_file(to_upload.read(), channel_name, dest)

        after_upload_move.append(dest)

    for f_to_move in after_upload_move:
        logger.debug(
            "Moving to final destination: "
            f"{f_to_move} -> {f_to_move[:-len(tmp_suffix)]}"
        )
        pkgstore.move_file(channel_name, f_to_move, f_to_move[: -len(tmp_suffix)])

    tempdir.cleanup()



"""new package versions

Revision ID: 8dfb7c4bfbd7
Revises: 98c04a65df4a
Create Date: 2020-12-21 16:35:33.940460

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '8dfb7c4bfbd7'
down_revision = '98c04a65df4a'
branch_labels = None
depends_on = None

enum_values = ('hour', 'day', 'month', 'year')


def get_intervaltype():
    if op.get_context().dialect.name == 'postgresql':
        intervaltype = sa.dialects.postgresql.ENUM(
            *enum_values, name='intervaltype', create_type=False
        )
    else:
        intervaltype = sa.Enum(*enum_values, name='intervaltype')  # type: ignore
    return intervaltype


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    intervaltype = get_intervaltype()

    op.drop_table('package_version_metrics')
    op.create_table(
        'aggregated_metrics',
        sa.Column('id', sa.LargeBinary(length=16), nullable=False),
        sa.Column('channel_name', sa.String(), nullable=True),
        sa.Column('platform', sa.String(), nullable=True),
        sa.Column('filename', sa.String(), nullable=True),
        sa.Column('metric_name', sa.String(length=255), nullable=False),
        sa.Column('period', intervaltype, nullable=True),
        sa.Column('count', sa.Integer(), server_default=sa.text('0'), nullable=False),
        sa.Column('timestamp', sa.DateTime(), nullable=False),
        sa.PrimaryKeyConstraint('id'),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    intervaltype = get_intervaltype()

    op.create_table(
        'package_version_metrics',
        sa.Column(
            'package_version_id',
            postgresql.BYTEA(),
            autoincrement=False,
            nullable=False,
        ),
        sa.Column(
            'metric_name', sa.VARCHAR(length=255), autoincrement=False, nullable=False
        ),
        sa.Column('period', intervaltype, autoincrement=False, nullable=False),
        sa.Column(
            'count',
            sa.INTEGER(),
            server_default=sa.text('0'),
            autoincrement=False,
            nullable=False,
        ),
        sa.Column(
            'timestamp', postgresql.TIMESTAMP(), autoincrement=False, nullable=False
        ),
        sa.ForeignKeyConstraint(
            ['package_version_id'],
            ['package_versions.id'],
            name='package_version_metrics_package_version_id_fkey',
        ),
        sa.PrimaryKeyConstraint(
            'package_version_id',
            'metric_name',
            'period',
            'timestamp',
            name='package_version_metrics_pkey',
        ),
    )
    op.drop_table('aggregated_metrics')
    # ### end Alembic commands ###



import logging
from typing import Dict

from . import BaseAuthenticator

logger = logging.getLogger("quetz")


class AuthenticatorRegistry:
    _instance = None
    _router = None
    enabled_authenticators: Dict[str, BaseAuthenticator] = {}

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    @classmethod
    def set_router(cls, router):
        cls._router = router

    def register(self, auth: BaseAuthenticator):
        if auth.provider in self.enabled_authenticators:
            logger.warning(f"authenticator '{auth.provider}' already registered")
            return

        if not self._router:
            raise Exception(
                "AuthenticationRegistry not completely configure, you need to set the"
                "root router using set_router method"
            )
        self._router.include_router(auth.router)
        self.enabled_authenticators[auth.provider] = auth
        logger.info(
            f"authentication provider '{auth.provider}' "
            f"of class {auth.__class__.__name__} registered"
        )

        if len(self.enabled_authenticators) > 1:
            logger.warning(
                "You have registered multiple authentication providers."
                "Please note that this is currently discouraged in production setups!"
            )

    def is_registered(self, provider_name: str):
        return provider_name in self.enabled_authenticators



import os
import shutil
import uuid
from datetime import date
from unittest import mock
from urllib.parse import quote

from pytest import fixture

from quetz.authorization import Rules
from quetz.db_models import (
    ApiKey,
    Channel,
    ChannelMember,
    Package,
    PackageMember,
    PackageVersion,
    Profile,
    User,
)


class Data:
    def __init__(self, db):
        self.keya = "akey"
        self.keyb = "bkey"

        self.usera = User(id=uuid.uuid4().bytes, username='usera')
        Profile(name='usera', user=self.usera, avatar_url='')
        db.add(self.usera)

        self.userb = User(id=uuid.uuid4().bytes, username='userb')
        Profile(name='userb', user=self.userb, avatar_url='')
        db.add(self.userb)

        self.userc = User(id=uuid.uuid4().bytes, username='userc', role="owner")
        Profile(name='userc', user=self.userc, avatar_url='')
        db.add(self.userc)

        assert len(db.query(User).all()) == 3

        self.keya_obj = ApiKey(
            key=self.keya,
            time_created=date.today(),
            expire_at=date(2030, 1, 1),
            user_id=self.usera.id,
            owner_id=self.usera.id,
        )

        db.add(self.keya_obj)
        db.add(
            ApiKey(
                key=self.keyb,
                time_created=date.today(),
                expire_at=date(2030, 1, 1),
                user_id=self.userb.id,
                owner_id=self.userb.id,
            )
        )

        self.channel1 = Channel(name="testchannel", private=False)
        self.channel2 = Channel(name="privatechannel", private=True)

        self.package1 = Package(name="package1", channel=self.channel1)
        self.package2 = Package(name="package2", channel=self.channel2)

        self.package_version_1 = PackageVersion(
            id=uuid.uuid4().bytes,
            channel_name=self.channel1.name,
            package_name=self.package1.name,
            package_format='tarbz2',
            platform='noarch',
            version="0.0.1",
            build_number="0",
            build_string="",
            filename="filename.tar.bz2",
            info="{}",
            uploader_id=self.usera.id,
            size=101,
        )
        self.package_version_2 = PackageVersion(
            id=uuid.uuid4().bytes,
            channel_name=self.channel2.name,
            package_name=self.package2.name,
            package_format='tarbz2',
            platform='noarch',
            version="0.0.1",
            build_number="0",
            build_string="",
            filename="filename2.tar.bz2",
            info="{}",
            uploader_id=self.usera.id,
            size=101,
        )

        self.channel_member = ChannelMember(
            channel=self.channel2, user=self.usera, role='maintainer'
        )
        self.channel_member_userc = ChannelMember(
            channel=self.channel2, user=self.userc, role='owner'
        )

        self.package_member = PackageMember(
            channel=self.channel2, user=self.userc, package=self.package2, role="owner"
        )

        for el in [
            self.channel1,
            self.channel2,
            self.channel_member,
            self.channel_member_userc,
            self.package1,
            self.package2,
            self.package_version_1,
            self.package_version_2,
            self.package_member,
        ]:
            db.add(el)
        db.commit()

        self.db = db


@fixture
def channel_dirs(data, config):
    # need to use config to get the right workdir
    os.makedirs('channels/testchannel/noarch', exist_ok=True)
    os.makedirs('channels/privatechannel/noarch', exist_ok=True)

    with open("channels/testchannel/noarch/current_repodata.json", "a") as f:
        f.write("file content 0")
    with open("channels/privatechannel/noarch/current_repodata.json", "a") as f:
        f.write("file content 1")

    yield

    shutil.rmtree('channels')


@fixture
def data(db):
    data = Data(db)
    yield data


def test_private_channels(data, client):
    response = client.get('/')
    assert len(response.text)
    response = client.get('/api/channels')

    response = client.get('/api/channels', headers={"X-Api-Key": data.keya})
    assert len(response.json()) == 2
    assert {c['name'] for c in response.json()} == {
        c.name for c in [data.channel1, data.channel2]
    }

    response = client.get('/api/channels', headers={"X-Api-Key": data.keyb})
    assert len(response.json()) == 1
    assert response.json()[0]['name'] == data.channel1.name

    response = client.get('/api/channels')
    assert len(response.json()) == 1
    assert response.json()[0]['name'] == data.channel1.name

    # Channel #

    # public access to public channel
    response = client.get(f'/api/channels/{data.channel1.name}')
    assert response.status_code == 200
    assert ('name', data.channel1.name) in response.json().items()

    # public access to private channel
    response = client.get(f'/api/channels/{data.channel2.name}')
    assert response.status_code == 401

    # non-member credential access to private channel
    response = client.get(
        f'/api/channels/{data.channel2.name}', headers={"X-Api-Key": data.keyb}
    )
    assert response.status_code == 403

    # member credential access to private channel
    response = client.get(
        f'/api/channels/{data.channel2.name}', headers={"X-Api-Key": data.keya}
    )
    assert response.status_code == 200
    assert ('name', data.channel2.name) in response.json().items()

    # Packages #

    # public access to public channel
    response = client.get(f'/api/channels/{data.channel1.name}/packages')
    assert response.status_code == 200
    assert len(response.json()) == 1
    assert ('name', data.package1.name) in response.json()[0].items()

    # public access to private channel
    response = client.get(f'/api/channels/{data.channel2.name}/packages')
    assert response.status_code == 401

    # non-member credential access to private channel
    response = client.get(
        f'/api/channels/{data.channel2.name}/packages', headers={"X-Api-Key": data.keyb}
    )
    assert response.status_code == 403

    # member credential access to private channel
    response = client.get(
        f'/api/channels/{data.channel2.name}/packages', headers={"X-Api-Key": data.keya}
    )
    assert response.status_code == 200
    assert len(response.json()) == 1
    assert ('name', data.package2.name) in response.json()[0].items()

    # Channel Members #

    # public access to public channel
    response = client.get(f'/api/channels/{data.channel1.name}/members')
    assert response.status_code == 401

    # public access to private channel
    response = client.get(f'/api/channels/{data.channel2.name}/members')
    assert response.status_code == 401

    # non-member credential access to private channel
    response = client.get(
        f'/api/channels/{data.channel2.name}/members', headers={"X-Api-Key": data.keyb}
    )
    assert response.status_code == 403

    # member credential access to private channel
    response = client.get(
        f'/api/channels/{data.channel2.name}/members', headers={"X-Api-Key": data.keya}
    )
    assert response.status_code == 200
    assert len(response.json()) == 2
    assert ('role', data.channel_member.role) in response.json()[0].items()
    assert response.json()[0]['user']['id'] == str(uuid.UUID(bytes=data.usera.id))

    # Package #

    # public access to public channel
    response = client.get(
        f'/api/channels/{data.channel1.name}/packages/{data.package1.name}'
    )
    assert response.status_code == 200
    assert ('name', data.package1.name) in response.json().items()

    # public access to private channel
    response = client.get(
        f'/api/channels/{data.channel2.name}/packages/{data.package2.name}'
    )
    assert response.status_code == 401

    # non-member credential access to private channel
    response = client.get(
        f'/api/channels/{data.channel2.name}/packages/{data.package2.name}',
        headers={"X-Api-Key": data.keyb},
    )
    assert response.status_code == 403

    # member credential access to private channel
    response = client.get(
        f'/api/channels/{data.channel2.name}/packages/{data.package2.name}',
        headers={"X-Api-Key": data.keya},
    )
    assert response.status_code == 200
    assert ('name', data.package2.name) in response.json().items()

    # package Members #

    # public access to public channel
    response = client.get(
        f'/api/channels/{data.channel1.name}/packages/{data.package1.name}/members'
    )
    assert response.status_code == 200
    assert len(response.json()) == 0

    # public access to private channel
    response = client.get(
        f'/api/channels/{data.channel2.name}/packages/{data.package2.name}/members'
    )
    assert response.status_code == 401

    # non-member credential access to private channel
    response = client.get(
        f'/api/channels/{data.channel2.name}/packages/{data.package2.name}/members',
        headers={"X-Api-Key": data.keyb},
    )
    assert response.status_code == 403

    # member credential access to private channel
    response = client.get(
        f'/api/channels/{data.channel2.name}/packages/{data.package2.name}/members',
        headers={"X-Api-Key": data.keya},
    )
    assert response.status_code == 200
    assert len(response.json()) == 1

    # package Versions #

    # public access to public channel
    response = client.get(
        f'/api/channels/{data.channel1.name}/packages/{data.package1.name}/versions'
    )
    assert response.status_code == 200
    assert len(response.json()) == 1

    # public access to private channel
    response = client.get(
        f'/api/channels/{data.channel2.name}/packages/{data.package2.name}/versions'
    )
    assert response.status_code == 401

    # non-member credential access to private channel
    response = client.get(
        f'/api/channels/{data.channel2.name}/packages/{data.package2.name}/versions',
        headers={"X-Api-Key": data.keyb},
    )
    assert response.status_code == 403

    # member credential access to private channel
    response = client.get(
        f'/api/channels/{data.channel2.name}/packages/{data.package2.name}/versions',
        headers={"X-Api-Key": data.keya},
    )
    assert response.status_code == 200
    assert len(response.json()) == 1

    # Package Search #
    # query = f"channel:test -format:conda uploader:{data.usera.username}"
    query = "channel:test"
    query = quote(query)
    response = client.get(f'/api/packages/search/?q={query}')
    assert response.status_code == 200
    assert len(response.json()) == 1
    assert response.json()[0]['name'] == data.package1.name

    # query = f"format:conda platform:linux-64,noarch uploader:{data.userb.username}"
    # query = quote(query)
    # response = client.get(
    #     f'/api/packages/search/?q={query}', headers={"X-Api-Key": data.keyb}
    # )
    # assert response.status_code == 200
    # assert len(response.json()) == 0

    # query = f"-channel:{data.channel2.name} format:tarbz2,conda -platform:osx-64"
    query = f"-channel:{data.channel2.name}"
    query = quote(query)
    response = client.get(
        f'/api/packages/search/?q={query}', headers={"X-Api-Key": data.keyb}
    )
    assert response.status_code == 200
    assert len(response.json()) == 1
    assert response.json()[0]['name'] == data.package1.name

    query = f"package NOT frame channel:{data.channel1.name},private"
    query = quote(query)
    response = client.get(
        f'/api/packages/search/?q={query}', headers={"X-Api-Key": data.keya}
    )
    assert response.status_code == 200
    assert len(response.json()) == 2
    assert {c['name'] for c in response.json()} == {
        c.name for c in [data.package1, data.package2]
    }

    # Channel Search #
    query = "test private:no"
    query = quote(query)
    response = client.get(f'/api/channels/search/?q={query}')
    assert response.status_code == 200
    assert len(response.json()) == 1
    assert response.json()[0]['name'] == data.channel1.name

    query = "private:y"
    query = quote(query)
    response = client.get(f'/api/channels/search/?q={query}')
    assert response.status_code == 200
    assert len(response.json()) == 0

    query = "-private:false"
    query = quote(query)
    response = client.get(
        f'/api/channels/search/?q={query}', headers={"X-Api-Key": data.keya}
    )
    assert response.status_code == 200
    assert len(response.json()) == 1
    assert response.json()[0]['name'] == data.channel2.name


def test_private_channels_create_package(data, client):
    # public access to public channel
    response = client.post(
        f'/api/channels/{data.channel1.name}/packages', json={"name": "newpackage1"}
    )
    assert response.status_code == 401

    # public access to private channel
    response = client.post(
        f'/api/channels/{data.channel2.name}/packages', json={"name": "newpackage1"}
    )
    assert response.status_code == 401

    # channel member can not create packages in public channel
    response = client.post(
        f'/api/channels/{data.channel1.name}/packages',
        json={"name": "newpackage2"},
        headers={"X-Api-Key": data.keyb},
    )
    assert response.status_code == 403

    # user with credentials to private channel
    response = client.post(
        f'/api/channels/{data.channel2.name}/packages',
        json={"name": "newpackage2"},
        headers={"X-Api-Key": data.keyb},
    )
    assert response.status_code == 403

    # member credential access to private channel
    response = client.post(
        f'/api/channels/{data.channel2.name}/packages',
        json={"name": "newpackage2"},
        headers={"X-Api-Key": data.keya},
    )
    assert response.status_code == 201


def test_private_channels_download(db, client, data, channel_dirs):
    # succeed on public channel
    response = client.get('/get/testchannel/noarch/current_repodata.json')
    assert response.status_code == 200
    assert response.text == "file content 0"

    # keep functioning when unnecessary token is provided
    response = client.get('/t/[api-key]/get/testchannel/noarch/current_repodata.json')
    assert response.status_code == 200
    assert response.text == "file content 0"

    # fail on private channel without credentials
    response = client.get('/get/privatechannel/noarch/current_repodata.json')
    assert response.status_code == 401

    # fail on private channel with invalid credentials
    response = client.get(
        '/t/[invalid-api-key]/get/privatechannel/noarch/current_repodata.json'
    )
    assert response.status_code == 401

    # fail on private channel with non member user
    response = client.get(
        f'/t/{data.keyb}/get/privatechannel/noarch/current_repodata.json'
    )
    assert response.status_code == 403

    # succeed on private channel with member user
    response = client.get(
        f'/t/{data.keya}/get/privatechannel/noarch/current_repodata.json'
    )
    assert response.status_code == 200
    assert response.text == "file content 1"


def test_create_api_key(data, client):
    response = client.get(f"/api/dummylogin/{data.userc.username}")
    assert response.status_code == 200

    response = client.post(
        '/api/api-keys',
        json={
            "description": "test-key",
            "roles": [{"channel": "privatechannel", "role": "member"}],
        },
    )

    assert response.status_code == 201
    assert response.json() == {
        "description": "test-key",
        'time_created': mock.ANY,
        'expire_at': None,
        "roles": [{"channel": "privatechannel", "role": "member", "package": None}],
        "key": mock.ANY,
    }

    # get key with package permissions

    response = client.post(
        '/api/api-keys',
        json={
            "description": "test-key",
            'expire_at': '2024-01-26',
            "roles": [
                {
                    "channel": "privatechannel",
                    "package": data.package2.name,
                    "role": "member",
                }
            ],
        },
    )

    assert response.status_code == 201
    assert response.json() == {
        "description": "test-key",
        'time_created': mock.ANY,
        'expire_at': '2024-01-26',
        "roles": [
            {
                "role": "member",
                "package": data.package2.name,
                "channel": "privatechannel",
            }
        ],
        "key": mock.ANY,
    }

    # get a key with user privileges

    response = client.post(
        '/api/api-keys',
        json={
            "description": "test-key",
            'expire_at': '2024-01-26',
            "roles": [],
        },
    )

    assert response.status_code == 201
    assert response.json() == {
        "description": "test-key",
        'time_created': mock.ANY,
        'expire_at': '2024-01-26',
        "roles": None,
        "key": mock.ANY,
    }


def test_use_wildcard_api_key_to_authenticate(data, client):
    response = client.get(f"/api/dummylogin/{data.userc.username}")
    assert response.status_code == 200

    response = client.post(
        '/api/api-keys',
        json={
            "description": "test-key",
            'expire_at': '2030-01-01',
            "roles": [],
        },
    )

    assert response.status_code == 201

    key = response.json()["key"]

    # per-channel key
    response = client.post(
        '/api/api-keys',
        json={
            "description": "test-key",
            'expire_at': '2030-01-01',
            "roles": [{"channel": "privatechannel", "role": "maintainer"}],
        },
    )

    assert response.status_code == 201
    channel_key = response.json()["key"]

    # add a new channel

    response = client.post(
        "/api/channels", json={"name": "my-new-channel", "private": True}
    )

    assert response.status_code == 201

    # clear session cookies
    client.cookies.clear()

    response = client.get(
        "/api/channels/privatechannel/members", headers={"X-API-Key": key}
    )

    assert response.status_code == 200

    response = client.get(
        "/api/channels/privatechannel/packages", headers={"X-API-Key": key}
    )
    assert response.status_code == 200

    response = client.get(
        "/api/channels/my-new-channel/members", headers={"X-API-Key": key}
    )

    assert response.status_code == 200

    response = client.get(
        "/api/channels/my-new-channel/packages", headers={"X-API-Key": key}
    )
    assert response.status_code == 200

    # using per-channel key

    response = client.get(
        "/api/channels/privatechannel/members", headers={"X-API-Key": channel_key}
    )

    assert response.status_code == 200

    response = client.get(
        "/api/channels/privatechannel/packages", headers={"X-API-Key": channel_key}
    )
    assert response.status_code == 200

    response = client.get(
        "/api/channels/my-new-channel/members", headers={"X-API-Key": channel_key}
    )

    assert response.status_code == 403

    response = client.get(
        "/api/channels/my-new-channel/packages", headers={"X-API-Key": channel_key}
    )
    assert response.status_code == 403


def test_authorizations_with_deleted_api_key(data: Data, db):
    auth = Rules(data.keya, {}, db)

    user_id = auth.get_user()

    assert user_id == data.usera.id

    data.keya_obj.deleted = True

    db.commit()

    user_id = auth.get_user()

    assert not user_id

    data.keya_obj.deleted = False


def test_authorizations_with_expired_api_key(data, client):
    response = client.get(f"/api/dummylogin/{data.userc.username}")
    assert response.status_code == 200

    response = client.post(
        '/api/api-keys',
        json={
            "description": "test-key",
            'expire_at': '2020-01-01',
            "roles": [],
        },
    )

    assert response.status_code == 201

    key = response.json()["key"]

    # clear session cookies
    client.cookies.clear()

    response = client.get(
        "/api/channels/privatechannel/members", headers={"X-API-Key": key}
    )

    assert response.status_code == 401


def test_authorizations_with_api_key_no_expiry(data, client, db):
    key = "noexpirykey"
    keya_obj = ApiKey(
        key=key,
        time_created=date.today(),
        expire_at=None,
        user_id=data.usera.id,
        owner_id=data.usera.id,
    )
    db.add(keya_obj)
    db.commit()

    response = client.get(
        f"/api/channels/{data.channel2.name}/members", headers={"X-API-Key": key}
    )

    assert response.status_code == 200



import grp
import logging
import os
import pwd
from typing import List, Optional

from fastapi import Request

from quetz.authentication.base import BaseAuthenticator, FormHandlers, UserProfile
from quetz.authorization import ServerRole
from quetz.config import Config, ConfigEntry, ConfigSection

logger = logging.getLogger("quetz")

try:
    import pamela

    class PAMAuthenticator(BaseAuthenticator):
        """Use PAM to authenticate with local system users.

        To enable add the following to your configuration file:

        .. code-block::

          [pamauthenticator]
          # name for the provider, used in the login URL
          provider = "pam"
          # use the following to translate the Unix groups that
          # users might belong to user role on Quetz server
          admin_groups = ["root", "quetz"]
          maintainer_groups = []
          user_groups = []

        On most Linux systems you can add users with:

        .. code-block::

          useradd USERNAME
          # set password interactively with
          passwd USERNAME

        .. note::

          For this authenticator to work, the user who runs the server must be root or
          be in ``shadow`` group.
        """

        provider: str = 'pam'
        handler_cls = FormHandlers

        service: str = "login"
        encoding: str = 'utf8'
        check_account: bool = True

        # configure server roles
        admin_groups: List[str] = []
        maintainer_groups: List[str] = []
        member_groups: List[str] = []

        def _make_config(self):
            section = ConfigSection(
                "pamauthenticator",
                [
                    ConfigEntry("provider", str, default="pam", required=False),
                    ConfigEntry("service", str, default="login", required=False),
                    ConfigEntry("encoding", str, default="utf8", required=False),
                    ConfigEntry("check_account", bool, default=True, required=False),
                    ConfigEntry("admin_groups", list, default=list, required=False),
                    ConfigEntry(
                        "maintainer_groups", list, default=list, required=False
                    ),
                    ConfigEntry("member_groups", list, default=list, required=False),
                ],
            )
            return [section]

        def _get_group_id_by_name(self, groupname):
            return grp.getgrnam(groupname).gr_gid

        def _get_user_gid_by_name(self, username):
            return pwd.getpwnam(username).pw_gid

        def _get_group_ids(self, group_names):
            gids = []
            for group in group_names:
                try:
                    gids.append(self._get_group_id_by_name(group))
                except Exception as exc:
                    logger.warning(f"got error {exc}.  Group {group} may not exist")
            return gids

        def _get_user_group_ids(self, username):
            user_gid = self._get_user_gid_by_name(username)
            return os.getgrouplist(username, user_gid)

        def configure(self, config: Config):
            config_options = self._make_config()
            config.register(config_options)

            if config.configured_section("pamauthenticator"):
                self.provider = config.pamauthenticator_provider
                self.service = config.pamauthenticator_service
                self.encoding = config.pamauthenticator_encoding
                self.check_account = config.pamauthenticator_check_account
                self.admin_groups = config.pamauthenticator_admin_groups
                self.maintainer_groups = config.pamauthenticator_maintainer_groups
                self.member_groups = config.pamauthenticator_member_groups
                self.is_enabled = True
            else:
                self.is_enabled = False

            super().configure(config)

        async def user_role(self, request: Request, profile: UserProfile):
            mappings = [
                (ServerRole.OWNER, self.admin_groups),
                (ServerRole.MAINTAINER, self.maintainer_groups),
                (ServerRole.MEMBER, self.member_groups),
            ]
            username = profile["login"]

            user_gids = self._get_user_group_ids(username)

            for role, groups in mappings:
                role_gids = self._get_group_ids(groups)
                common = set(role_gids) & set(user_gids)
                if common:
                    logger.info(
                        f"pam authenticator: user {username} found in group {common}, "
                        f"setting {role} permissions"
                    )
                    return role.value

        async def authenticate(
            self,
            request: Request,
            data: Optional[dict] = None,
            dao=None,
            config=None,
            **kwargs,
        ) -> Optional[str]:
            """Authenticate with PAM, and return the username if login is successful.
            Return None otherwise.
            """
            if data is None:
                return None

            host = None
            if request.client:
                host = request.client.host

            username = data['username']
            try:
                pamela.authenticate(
                    username,
                    data['password'],
                    service=self.service,
                    encoding=self.encoding,
                )
            except pamela.PAMError as e:
                logger.warning(f"PAM Authentication failed ({username}@{host}): {e}")
                return None

            if self.check_account:
                try:
                    pamela.check_account(
                        username, service=self.service, encoding=self.encoding
                    )
                except pamela.PAMError as e:
                    logger.warning(f"PAM Account Check failed ({username}@{host}): {e}")
                    return None

            return username

except ImportError:
    PAMAuthenticator = None  # type: ignore



import hashlib
import os
import shutil
import time
import uuid
from collections.abc import Collection
from io import BytesIO
from pathlib import Path

import pytest

from quetz.pkgstores import (
    AzureBlobStore,
    GoogleCloudStorageStore,
    LocalStore,
    PackageStore,
    S3Store,
    has_xattr,
)

s3_config = {
    'key': os.environ.get("S3_ACCESS_KEY"),
    'secret': os.environ.get("S3_SECRET_KEY"),
    'url': os.environ.get("S3_ENDPOINT"),
    'region': os.environ.get("S3_REGION"),
    'bucket_prefix': "test",
    'bucket_suffix': "",
}

azure_config = {
    'account_name': os.environ.get("AZURE_ACCOUNT_NAME"),
    'account_access_key': os.environ.get("AZURE_ACCESS_KEY"),
    'conn_str': os.environ.get("AZURE_CONN_STRING"),
    'container_prefix': "test",
    'container_suffix': "",
}

gcs_config = {
    'project': os.environ.get("GCS_PROJECT"),
    'token': os.environ.get("GCS_TOKEN"),
    'bucket_prefix': "test",
    'bucket_suffix': "",
}

test_dir = os.path.dirname(__file__)


@pytest.mark.parametrize("redirect_enabled", [False, True])
def test_local_store(redirect_enabled):
    temp_dir = os.path.join(test_dir, "test_pkg_store_" + str(int(time.time())))
    os.makedirs(temp_dir, exist_ok=False)

    pkg_store = LocalStore(
        {
            "channels_dir": temp_dir,
            "redirect_enabled": redirect_enabled,
            "redirect_endpoint": "/files",
        }
    )

    pkg_store.add_file("content", "my-channel", "test.txt")
    pkg_store.add_file("content".encode('ascii'), "my-channel", "test_2.txt")

    files = pkg_store.list_files("my-channel")

    assert files == ["test.txt", "test_2.txt"]

    pkg_store.delete_file("my-channel", "test.txt")

    files = pkg_store.list_files("my-channel")
    assert files == ["test_2.txt"]

    with pkg_store.serve_path("my-channel", "test_2.txt") as f:
        assert f.read().decode('utf-8') == "content"

    metadata = pkg_store.get_filemetadata("my-channel", "test_2.txt")

    assert metadata[0] > 0
    assert type(metadata[1]) is float

    if has_xattr:
        md5 = hashlib.md5("content".encode('ascii')).hexdigest()
        assert metadata[2] == md5
    else:
        assert type(metadata[2]) is str

    shutil.rmtree(temp_dir)


@pytest.fixture
def channel_name():
    return "mychannel" + str(uuid.uuid4())


@pytest.fixture
def local_store():
    temp_dir = os.path.join(test_dir, "test_pkg_store_" + str(int(time.time())))
    os.makedirs(temp_dir, exist_ok=False)

    pkg_store = LocalStore(
        {
            "channels_dir": temp_dir,
            "redirect_enabled": False,
            "redirect_endpoint": "/files",
        }
    )

    yield pkg_store

    shutil.rmtree(temp_dir)


@pytest.fixture(
    params=[
        "local_store",
        pytest.param(
            "s3_store",
            marks=pytest.mark.skipif(
                not s3_config['key'], reason="requires s3 credentials"
            ),
        ),
        pytest.param(
            "azure_store",
            marks=pytest.mark.skipif(
                not azure_config['account_access_key'],
                reason="requires azure credentials",
            ),
        ),
        pytest.param(
            "gcs_store",
            marks=pytest.mark.skipif(
                not gcs_config['project'],
                reason="requires google cloud storage credentials",
            ),
        ),
    ]
)
def any_store(request):
    val = request.getfixturevalue(request.param)
    return val


def test_remove_dirs(any_store, channel_name):
    any_store.create_channel(channel_name)
    any_store.add_file("content", channel_name, "test.txt")

    with any_store.serve_path(channel_name, "test.txt") as f:
        assert f.read().decode('utf-8') == "content"

    any_store.remove_channel(channel_name)

    with pytest.raises(FileNotFoundError):
        with any_store.serve_path(channel_name, "test.txt") as f:
            f.read()


@pytest.fixture
def s3_store():
    pkg_store = S3Store(s3_config)
    return pkg_store


@pytest.fixture
def azure_store():
    pkg_store = AzureBlobStore(azure_config)

    yield pkg_store


@pytest.fixture
def gcs_store():
    pkg_store = GoogleCloudStorageStore(gcs_config)

    yield pkg_store


@pytest.fixture
def channel(any_store, channel_name):
    any_store.create_channel(channel_name)

    yield

    # cleanup
    any_store.remove_channel(channel_name)


def assert_files(
    pkg_store: PackageStore,
    channel_name: str,
    expected_files: Collection[str],
    n_retries: int = 3,
):
    """
    Asserts that the files in the package store match the expected files with retry.
    """
    for _ in range(n_retries):
        files = pkg_store.list_files(channel_name)
        try:
            assert files == expected_files
        except AssertionError:
            continue
        else:
            assert set(files) == set(expected_files)


def test_store_add_list_files(any_store, channel, channel_name):
    pkg_store = any_store

    pkg_store.add_file("content", channel_name, "test.txt")
    pkg_store.add_file("content", channel_name, "test_2.txt")

    assert_files(pkg_store, channel_name, ["test.txt", "test_2.txt"])

    pkg_store.delete_file(channel_name, "test.txt")

    assert_files(pkg_store, channel_name, ["test_2.txt"])

    metadata = pkg_store.get_filemetadata(channel_name, "test_2.txt")
    assert metadata[0] > 0
    assert type(metadata[1]) is float
    assert type(metadata[2]) is str


@pytest.mark.asyncio
async def test_add_package_async(any_store, channel, channel_name):
    pkg_store = any_store

    data = (Path(__file__).parent / "data" / "test-package-0.1-0.tar.bz2").read_bytes()

    await pkg_store.add_package_async(
        BytesIO(data), channel_name, "test-package-0.1-0.tar.gz"
    )

    assert_files(pkg_store, channel_name, ["test-package-0.1-0.tar.gz"])


def test_add_package(any_store, channel, channel_name):
    pkg_store = any_store

    data = (Path(__file__).parent / "data" / "test-package-0.1-0.tar.bz2").read_bytes()

    pkg_store.add_package(BytesIO(data), channel_name, "test-package-0.1-0.tar.gz")

    assert_files(pkg_store, channel_name, ["test-package-0.1-0.tar.gz"])


def test_move_file(any_store, channel, channel_name):
    pkg_store = any_store

    pkg_store.add_file("content", channel_name, "test.txt")
    pkg_store.move_file(channel_name, "test.txt", "test_2.txt")

    assert_files(pkg_store, channel_name, ['test_2.txt'])


@pytest.mark.parametrize("redirect_enabled", [False, True])
@pytest.mark.parametrize("redirect_endpoint", ["/files", "/static"])
def test_local_store_url(redirect_enabled, redirect_endpoint):
    local_store = LocalStore(
        {
            "channels_dir": "channels",
            "redirect_enabled": redirect_enabled,
            "redirect_endpoint": redirect_endpoint,
        }
    )
    url = local_store.url("mychannel", "linux-64/foo.tar.bz2")
    if redirect_enabled:
        assert url == f"{redirect_endpoint}/channels/mychannel/linux-64/foo.tar.bz2"
    else:
        assert url == os.path.abspath("channels/mychannel/linux-64/foo.tar.bz2")


def test_create_channel_multiple_times(any_store, channel_name):
    any_store.create_channel(channel_name)
    any_store.create_channel(channel_name)
